import mlflow.models 
from mlflow import models

evaluate = mlflow.models.evaluate
Model = models.Model
FlavorBackend = mlflow.models.FlavorBackend
infer_pip_requirements =  mlflow.models.infer_pip_requirements
evaluate = mlflow.models.evaluate
EvaluationArtifact = mlflow.models.EvaluationArtifact
EvaluationResult = mlflow.models.EvaluationResult
list_evaluators =  mlflow.models.list_evaluators

try:
    ModelSignature = mlflow.models.ModelSignature
    ModelInputExample = mlflow.models.ModelInputExample
    infer_signature = mlflow.models.infer_signature
    validate_schema = mlflow.models.validate_schema
except:
    pass


import mlflow.tracking._model_registry.fluent
import mlflow.tracking.fluent
from mlflow import tracking
from mlflow.tracking.request_header.abstract_request_header_provider import RequestHeaderProvider
from mlflow.entities import Experiment
from newron.auth import Auth0
import requests

get_tracking_uri = tracking.get_tracking_uri
get_registry_uri = tracking.get_registry_uri
create_experiment = mlflow.tracking.fluent.create_experiment
set_experiment = mlflow.tracking.fluent.set_experiment
log_params = mlflow.tracking.fluent.log_params
log_metrics = mlflow.tracking.fluent.log_metrics
set_experiment_tags = mlflow.tracking.fluent.set_experiment_tags
set_experiment_tag = mlflow.tracking.fluent.set_experiment_tag
set_tags = mlflow.tracking.fluent.set_tags
delete_experiment = mlflow.tracking.fluent.delete_experiment
delete_run = mlflow.tracking.fluent.delete_run
register_model = mlflow.tracking._model_registry.fluent.register_model
autolog = mlflow.tracking.fluent.autolog
evaluate = mlflow.models.evaluate
last_active_run = mlflow.tracking.fluent.last_active_run
NewronClient = tracking.MlflowClient
ActiveRun = mlflow.tracking.fluent.ActiveRun
log_param = mlflow.tracking.fluent.log_param
log_metric = mlflow.tracking.fluent.log_metric
set_tag = mlflow.tracking.fluent.set_tag
delete_tag = mlflow.tracking.fluent.delete_tag
log_artifacts = mlflow.tracking.fluent.log_artifacts
log_artifact = mlflow.tracking.fluent.log_artifact
log_text = mlflow.tracking.fluent.log_text
log_dict = mlflow.tracking.fluent.log_dict
log_image = mlflow.tracking.fluent.log_image
log_figure = mlflow.tracking.fluent.log_figure
active_run = mlflow.tracking.fluent.active_run
get_run = mlflow.tracking.fluent.get_run
start_run = mlflow.tracking.fluent.start_run
end_run = mlflow.tracking.fluent.end_run
search_runs = mlflow.tracking.fluent.search_runs
list_run_infos = mlflow.tracking.fluent.list_run_infos
get_artifact_uri = mlflow.tracking.fluent.get_artifact_uri
set_tracking_uri = tracking.set_tracking_uri
set_registry_uri = tracking.set_registry_uri
get_experiment = mlflow.tracking.fluent.get_experiment
get_experiment_by_name = mlflow.tracking.fluent.get_experiment_by_name
list_experiments = mlflow.tracking.fluent.list_experiments

SERVER_URI = "https://mlflow-tracking-server-zx44gn5asa-uc.a.run.app"
PROJECT_URI = "https://api.newron.ai/v1/set/project"
DEFAULT_AWAIT_MAX_SLEEP_SECONDS = 5 * 60

class NewronPluginRequestHeaderProvider(RequestHeaderProvider):
    """RequestHeaderProvider provided through plugin system"""

    def in_context(self):
        return True

    def request_headers(self):
        return {"project_id": Experiment.experiment_id}

def init(experiment_name, project_name, description=None):
    """
    Function to initialise a tracking experiment with Newron. The function authenticates
    the user against the Newron server and allows the user to activate an experiment under 
    a project. 

    Args:
        experiment_name: Case sensitive name of the experiment to be activated. If an experiment
                            with this name does not exist, a new experiment wth this name is
                            created.
        project_name: Case sensitive name of the project under which the experimentis
                            to be activated.
        description: Description of the experiment being activated. In case the experiment had a 
                            description previously it would be overwritten by the new description.
    """
    _auth = Auth0()
    auth_response = _auth.authenticate()
    if auth_response:
        set_tracking_uri(SERVER_URI)
        import requests
        url = "https://grpc-api-gateway-7boevord.uc.gateway.dev/v1/project"

        payload = {}
        payload["accountId"] = auth_response["email"]
        payload["userId"] = auth_response["sub"].split("|")[1]
        payload["projectName"] = project_name
        payload["experimentName"] = experiment_name
        if description:
            payload["desc"] = description

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + auth_response["access_token"]
        }
        gateway_response = requests.request("PUT", url, json=payload, headers=headers)
        set_experiment(experiment_id = gateway_response.json()["mlflow"]["experimentId"])
    else:
        raise Exception("Authentication failed")



import mlflow.spacy
import inspect
from mlflow.utils.environment import (
    _mlflow_conda_env,
    _validate_env_arguments,
    _process_pip_requirements,
    _process_conda_env,
    _CONDA_ENV_FILE_NAME,
    _REQUIREMENTS_FILE_NAME,
    _CONSTRAINTS_FILE_NAME,
    _PYTHON_ENV_FILE_NAME,
    _PythonEnv,
)
from mlflow.utils.requirements_utils import _get_pinned_requirement
from newron.models import ModelSignature, ModelInputExample
from mlflow.utils.model_utils import (
    _get_flavor_configuration,
    _validate_and_copy_code_paths,
    _add_code_from_conf_to_system_path,
    _validate_and_prepare_target_save_path,
)
FLAVOR_NAME = "spacy"
DEFAULT_AWAIT_MAX_SLEEP_SECONDS = 5 * 60

_load_model = mlflow.spacy._load_model
_SpacyModelWrapper = mlflow.spacy._SpacyModelWrapper
_load_pyfunc = mlflow.spacy._load_pyfunc


def get_default_pip_requirements():
    """
    
    Returns:
        A list of default pip requirements for MLflow Models produced by this flavor.
             Calls to :func:`save_model()` and :func:`log_model()` produce a pip environment
             that, at minimum, contains these requirements.
    """
    return [_get_pinned_requirement("spacy")]


def get_default_conda_env():
    """
    
    Returns:
        The default Conda environment for MLflow Models produced by calls to
             :func:`save_model()` and :func:`log_model()`.
    """
    return _mlflow_conda_env(additional_pip_deps=get_default_pip_requirements())

def save_model(
    spacy_model,
    path,
    conda_env=None,
    code_paths=None,
    mlflow_model=None,
    signature: ModelSignature = None,
    input_example: ModelInputExample = None,
    pip_requirements=None,
    extra_pip_requirements=None,
):
    """
    Save a spaCy model to a path on the local file system.

    Args:
        spacy_model: spaCy model to be saved.
        path: Local path where the model is to be saved.
        conda_env: conda_env
        code_paths: A list of local filesystem paths to Python file dependencies (or directories
                       containing file dependencies). These files are *prepended* to the system
                       path when the model is loaded.
        mlflow_model: :py:mod:`mlflow.models.Model` this flavor is being added to.
        signature: :py:class:`ModelSignature <mlflow.models.ModelSignature>`
                      describes model input and output :py:class:`Schema <mlflow.types.Schema>`.
                      The model signature can be :py:func:`inferred <mlflow.models.infer_signature>`
                      from datasets with valid model input (e.g. the training dataset with target
                      column omitted) and valid model output (e.g. model predictions generated on
                      the training dataset), for example:
                      .. code-block:: python
                        from mlflow.models.signature import infer_signature
                        train = df.drop_column("target_label")
                        predictions = ... # compute model predictions
                        signature = infer_signature(train, predictions)
        input_example: Input example provides one or several instances of valid
                          model input. The example can be used as a hint of what data to feed the
                          model. The given example will be converted to a Pandas DataFrame and then
                          serialized to json using the Pandas split-oriented format. Bytes are
                          base64-encoded.
        pip_requirements: pip_requirements
        extra_pip_requirements: extra_pip_requirements
    """
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    del values["frame"]
    mlflow.spacy.save_model(**values)

def log_model(
    spacy_model,
    artifact_path,
    conda_env=None,
    code_paths=None,
    registered_model_name=None,
    signature: ModelSignature = None,
    input_example: ModelInputExample = None,
    pip_requirements=None,
    extra_pip_requirements=None,
    **kwargs,
):
    """
    Log a spaCy model as an MLflow artifact for the current run.

    Args:
        spacy_model: spaCy model to be saved.
        artifact_path: Run-relative artifact path.
        conda_env: conda_env
        code_paths: A list of local filesystem paths to Python file dependencies (or directories
                       containing file dependencies). These files are *prepended* to the system
                       path when the model is loaded.
        registered_model_name: If given, create a model version under
                                  ``registered_model_name``, also creating a registered model if one
                                  with the given name does not exist.
        signature: :py:class:`ModelSignature <mlflow.models.ModelSignature>`
                      describes model input and output :py:class:`Schema <mlflow.types.Schema>`.
                      The model signature can be :py:func:`inferred <mlflow.models.infer_signature>`
                      from datasets with valid model input (e.g. the training dataset with target
                      column omitted) and valid model output (e.g. model predictions generated on
                      the training dataset), for example:
                      .. code-block:: python
                        from mlflow.models.signature import infer_signature
                        train = df.drop_column("target_label")
                        predictions = ... # compute model predictions
                        signature = infer_signature(train, predictions)
        input_example: Input example provides one or several instances of valid
                          model input. The example can be used as a hint of what data to feed the
                          model. The given example will be converted to a Pandas DataFrame and then
                          serialized to json using the Pandas split-oriented format. Bytes are
                          base64-encoded.
        pip_requirements: pip_requirements
        extra_pip_requirements: extra_pip_requirements
        kwargs: kwargs to pass to ``spacy.save_model`` method.
        
    Returns:
        A :py:class:`ModelInfo <mlflow.models.model.ModelInfo>` instance that contains the
             metadata of the logged model.
    """
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    del values["frame"]
    del values["kwargs"]
    return mlflow.spacy.log_model(**values)

def load_model(model_uri, dst_path=None):
    """
    Load a spaCy model from a local file (if ``run_id`` is ``None``) or a run.
    Args:    
        model_uri: The location, in URI format, of the MLflow model. For example:
            * ``Users/me/path/to/local/model``
            * ``relative/path/to/local/model``
            * ``s3://my_bucket/path/to/model``
            * ``runs:/<mlflow_run_id>/run-relative/path/to/model``
            * ``models:/<model_name>/<model_version>``
            * ``models:/<model_name>/<stage>``
                      For more information about supported URI schemes, see
                      `Referencing Artifacts <https://www.mlflow.org/docs/latest/concepts.html#
                      artifact-locations>`.
        dst_path: The local filesystem path to which to download the model artifact.
                     This directory must already exist. If unspecified, a local output
                     path will be created.
                     
    Returns:
        A spaCy loaded model
    """
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    del values["frame"]
    return mlflow.spacy.load_model(**values)


import os
from mlflow.deployments import BaseDeploymentClient

f_deployment_name = "fake_deployment_name"
f_endpoint_name = "fake_endpoint_name"


class PluginDeploymentClient(BaseDeploymentClient):
    def create_deployment(self, name, model_uri, flavor=None, config=None, endpoint=None):
        if config and config.get("raiseError") == "True":
            raise RuntimeError("Error requested")
        return {"name": f_deployment_name, "flavor": flavor}

    def delete_deployment(self, name, config=None, endpoint=None):
        if config and config.get("raiseError") == "True":
            raise RuntimeError("Error requested")
        return None

    def update_deployment(self, name, model_uri=None, flavor=None, config=None, endpoint=None):
        return {"flavor": flavor}

    def list_deployments(self, endpoint=None):
        if os.environ.get("raiseError") == "True":
            raise RuntimeError("Error requested")
        return [{"name": f_deployment_name}]

    def get_deployment(self, name, endpoint=None):
        return {"key1": "val1", "key2": "val2"}

    def predict(self, deployment_name=None, df=None, endpoint=None):
        return "1"

    def explain(self, deployment_name=None, df=None, endpoint=None):
        return "1"

    def create_endpoint(self, name, config=None):
        if config and config.get("raiseError") == "True":
            raise RuntimeError("Error requested")
        return {"name": f_endpoint_name}

    def update_endpoint(self, endpoint, config=None):
        return None

    def delete_endpoint(self, endpoint):
        return None

    def list_endpoints(self):
        return [{"name": f_endpoint_name}]

    def get_endpoint(self, endpoint):
        return {"name": f_endpoint_name}


def run_local(name, model_uri, flavor=None, config=None):
    # pylint: disable-next=print-function
    print(
        "Deployed locally at the key {} using the model from {}. ".format(name, model_uri)
        + "It's flavor is {} and config is {}".format(flavor, config)
    )


def target_help():
    return "Target help is called"



import os
import pathlib
import yaml
from unittest import mock

import pytest

import mlflow
from mlflow.pipelines.utils import get_pipeline_config
from mlflow.pipelines.utils.tracking import get_pipeline_tracking_config, log_code_snapshot
from mlflow.utils.file_utils import path_to_local_file_uri, path_to_local_sqlite_uri

# pylint: disable=unused-import
from tests.pipelines.helper_functions import (
    enter_pipeline_example_directory,
    enter_test_pipeline_directory,
    list_all_artifacts,
)  # pylint: enable=unused-import


@pytest.mark.usefixtures("enter_test_pipeline_directory")
@pytest.mark.parametrize(
    "tracking_uri,artifact_location,experiment_name,experiment_id",
    [
        ("mysql://myhost:8000/test_uri", "test/artifact/location", "myexpname", "myexpid"),
        ("mysql://myhost:8000/test_uri", "test/artifact/location", None, "myexpid"),
        ("mysql://myhost:8000/test_uri", "test/artifact/location", "myexpname", None),
        (None, "test/artifact/location", "myexpname", "myexpid"),
        ("mysql://myhost:8000/test_uri", None, "myexpname", "myexpid"),
        (None, None, None, "myexpid"),
    ],
)
def test_get_pipeline_tracking_config_returns_expected_config(
    tracking_uri, artifact_location, experiment_name, experiment_id
):
    default_tracking_uri = path_to_local_sqlite_uri(
        path=str((pathlib.Path.cwd() / "metadata" / "mlflow" / "mlruns.db").resolve())
    )
    default_artifact_location = path_to_local_file_uri(
        path=str((pathlib.Path.cwd() / "metadata" / "mlflow" / "mlartifacts").resolve())
    )
    default_experiment_name = "sklearn_regression"  # equivalent to pipeline name

    profile_contents = {
        "experiment": {},
        "INGEST_DATA_LOCATION": None,
        "INGEST_SCORING_DATA_LOCATION": None,
        "SCORED_OUTPUT_DATA_LOCATION": None,
    }
    if tracking_uri is not None:
        profile_contents["experiment"]["tracking_uri"] = tracking_uri
    if artifact_location is not None:
        profile_contents["experiment"]["artifact_location"] = artifact_location
    if experiment_name is not None:
        profile_contents["experiment"]["name"] = experiment_name
    if experiment_id is not None:
        profile_contents["experiment"]["id"] = experiment_id

    profile_path = pathlib.Path.cwd() / "profiles" / "testprofile.yaml"
    with open(profile_path, "w") as f:
        yaml.safe_dump(profile_contents, f)

    pipeline_config = get_pipeline_config(pipeline_root_path=os.getcwd(), profile="testprofile")
    pipeline_tracking_config = get_pipeline_tracking_config(
        pipeline_root_path=os.getcwd(), pipeline_config=pipeline_config
    )
    assert pipeline_tracking_config.tracking_uri == (tracking_uri or default_tracking_uri)
    assert pipeline_tracking_config.artifact_location == (
        artifact_location or default_artifact_location
    )
    if experiment_name is not None:
        assert pipeline_tracking_config.experiment_name == experiment_name
    elif experiment_id is not None:
        assert pipeline_tracking_config.experiment_id == experiment_id
    elif experiment_id is None and experiment_name is None:
        assert pipeline_tracking_config.experiment_name == default_experiment_name


@pytest.mark.usefixtures("enter_test_pipeline_directory")
@pytest.mark.parametrize(
    "tracking_uri,artifact_location,experiment_name,experiment_id",
    [
        ("mysql://myhost:8000/test_uri", "test/artifact/location", "myexpname", "myexpid"),
        ("mysql://myhost:8000/test_uri", "test/artifact/location", None, "myexpid"),
        ("mysql://myhost:8000/test_uri", "test/artifact/location", "myexpname", None),
        (None, "test/artifact/location", "myexpname", "myexpid"),
        ("mysql://myhost:8000/test_uri", None, "myexpname", "myexpid"),
        (None, None, None, "myexpid"),
    ],
)
def test_get_pipeline_tracking_config_returns_expected_config_on_databricks(
    tracking_uri, artifact_location, experiment_name, experiment_id
):
    with mock.patch("mlflow.pipelines.utils.tracking.is_in_databricks_runtime", return_value=True):
        default_tracking_uri = "databricks"
        default_experiment_name = "sklearn_regression"  # equivalent to pipeline name

        profile_contents = {
            "experiment": {},
            "INGEST_DATA_LOCATION": None,
            "INGEST_SCORING_DATA_LOCATION": None,
            "SCORED_OUTPUT_DATA_LOCATION": None,
        }
        if tracking_uri is not None:
            profile_contents["experiment"]["tracking_uri"] = tracking_uri
        if artifact_location is not None:
            profile_contents["experiment"]["artifact_location"] = artifact_location
        if experiment_name is not None:
            profile_contents["experiment"]["name"] = experiment_name
        if experiment_id is not None:
            profile_contents["experiment"]["id"] = experiment_id

        profile_path = pathlib.Path.cwd() / "profiles" / "testprofile.yaml"
        with open(profile_path, "w") as f:
            yaml.safe_dump(profile_contents, f)

        pipeline_config = get_pipeline_config(pipeline_root_path=os.getcwd(), profile="testprofile")
        pipeline_tracking_config = get_pipeline_tracking_config(
            pipeline_root_path=os.getcwd(), pipeline_config=pipeline_config
        )
        assert pipeline_tracking_config.tracking_uri == (tracking_uri or default_tracking_uri)
        assert pipeline_tracking_config.artifact_location == artifact_location
        if experiment_name is not None:
            assert pipeline_tracking_config.experiment_name == experiment_name
        elif experiment_id is not None:
            assert pipeline_tracking_config.experiment_id == experiment_id
        elif experiment_id is None and experiment_name is None:
            assert pipeline_tracking_config.experiment_name == default_experiment_name


def test_log_code_snapshot(tmp_path: pathlib.Path):
    files = [
        "pipeline.yaml",
        "requirements.txt",
        "profiles/local.yaml",
        "steps/train.py",
        "runtime/pipeline.yaml",
    ]
    for f in files:
        path = tmp_path.joinpath(f)
        path.parent.mkdir(exist_ok=True, parents=True)
        path.write_text("")

    mlflow.set_experiment(experiment_id="0")
    pipeline_config = {"name": "fake_config", "dict": {"key": 123}}
    with mlflow.start_run() as run:
        log_code_snapshot(tmp_path, run.info.run_id, pipeline_config=pipeline_config)
        tracking_uri = mlflow.get_tracking_uri()

    artifacts = set(list_all_artifacts(tracking_uri, run.info.run_id))
    assert artifacts.issuperset(f"pipeline_snapshot/{f}" for f in files)



import os
import inspect
import shutil
import subprocess
from unittest import mock

import pytest

import newron
import mlflow
from mlflow.utils.file_utils import path_to_local_sqlite_uri

from tests.autologging.fixtures import enable_test_mode


@pytest.fixture
def reset_mock():
    cache = []

    def set_mock(obj, attr, mock):
        cache.append((obj, attr, getattr(obj, attr)))
        setattr(obj, attr, mock)

    yield set_mock

    for obj, attr, value in cache:
        setattr(obj, attr, value)
    cache[:] = []


@pytest.fixture(autouse=True)
def tracking_uri_mock(tmpdir, request):
    try:
        if "notrackingurimock" not in request.keywords:
            tracking_uri = path_to_local_sqlite_uri(os.path.join(tmpdir.strpath, "mlruns.sqlite"))
            newron.set_tracking_uri(tracking_uri)
            os.environ["MLFLOW_TRACKING_URI"] = tracking_uri
        yield tmpdir
    finally:
        newron.set_tracking_uri(None)
        if "notrackingurimock" not in request.keywords:
            del os.environ["MLFLOW_TRACKING_URI"]


@pytest.fixture(autouse=True, scope="session")
def enable_test_mode_by_default_for_autologging_integrations():
    """
    Run all MLflow tests in autologging test mode, ensuring that errors in autologging patch code
    are raised and detected. For more information about autologging test mode, see the docstring
    for :py:func:`mlflow.utils.autologging_utils._is_testing()`.
    """
    yield from enable_test_mode()


@pytest.fixture(autouse=True)
def clean_up_leaked_runs():
    """
    Certain test cases validate safety API behavior when runs are leaked. Leaked runs that
    are not cleaned up between test cases may result in cascading failures that are hard to
    debug. Accordingly, this fixture attempts to end any active runs it encounters and
    throws an exception (which reported as an additional error in the pytest execution output).
    """
    try:
        yield
        assert (
            not newron.active_run()
        ), "test case unexpectedly leaked a run. Run info: {}. Run data: {}".format(
            newron.active_run().info, newron.active_run().data
        )
    finally:
        while newron.active_run():
            newron.end_run()


def _called_in_save_model():
    for frame in inspect.stack()[::-1]:
        if frame.function == "save_model":
            return True
    return False


@pytest.fixture(autouse=True)
def prevent_infer_pip_requirements_fallback(request):
    """
    Prevents `mlflow.models.infer_pip_requirements` from falling back in `mlflow.*.save_model`
    unless explicitly disabled via `pytest.mark.allow_infer_pip_requirements_fallback`.
    """
    from mlflow.utils.environment import _INFER_PIP_REQUIREMENTS_FALLBACK_MESSAGE

    def new_exception(msg, *_, **__):
        if msg == _INFER_PIP_REQUIREMENTS_FALLBACK_MESSAGE and _called_in_save_model():
            raise Exception(
                "`mlflow.models.infer_pip_requirements` should not fall back in"
                "`mlflow.*.save_model` during test"
            )

    if "allow_infer_pip_requirements_fallback" not in request.keywords:
        with mock.patch("mlflow.utils.environment._logger.exception", new=new_exception):
            yield
    else:
        yield


@pytest.fixture(autouse=True, scope="module")
def clean_up_mlruns_directory(request):
    """
    Clean up an `mlruns` directory on each test module teardown on CI to save the disk space.
    """
    yield

    # Only run this fixture on CI.
    if "GITHUB_ACTIONS" not in os.environ:
        return

    mlruns_dir = os.path.join(request.config.rootpath, "mlruns")
    if os.path.exists(mlruns_dir):
        try:
            shutil.rmtree(mlruns_dir)
        except IOError:
            if os.name == "nt":
                raise
            # `shutil.rmtree` can't remove files owned by root in a docker container.
            subprocess.run(["sudo", "rm", "-rf", mlruns_dir], check=True)


@pytest.fixture
def mock_s3_bucket():
    """
    Creates a mock S3 bucket using moto

    :return: The name of the mock bucket
    """
    import boto3
    import moto

    with moto.mock_s3():
        bucket_name = "mock-bucket"
        s3_client = boto3.client("s3")
        s3_client.create_bucket(Bucket=bucket_name)
        yield bucket_name


import json
import mlflow.pipelines.utils.step as step_utils
import numpy as np
import pytest

from mlflow.exceptions import MlflowException
from mlflow.pipelines.utils.step import (
    display_html,
    get_merged_eval_metrics,
    get_pandas_data_profile,
)
from pandas import DataFrame
from unittest import mock


def test_display_html_raises_without_input():
    with pytest.raises(MlflowException, match="At least one HTML source must be provided"):
        display_html()


def test_display_html_opens_html_data():
    html_data = "<!DOCTYPE html><html><body><p>Hey</p></body></html>"
    with mock.patch(
        "mlflow.pipelines.utils.step.is_running_in_ipython_environment", return_value=True
    ):
        with mock.patch("IPython.display.display") as patched_display:
            display_html(html_data=html_data)
            patched_display.assert_called_once()


def test_display_html_opens_html_file(tmp_path):
    html_file = tmp_path / "test.html"
    html_file.write_text("<!DOCTYPE html><html><body><p>Hey</p></body></html>")
    with mock.patch("subprocess.run") as patched_subprocess, mock.patch(
        "shutil.which", return_value=True
    ):
        display_html(html_file_path=html_file)
        patched_subprocess.assert_called_once()


@pytest.mark.parametrize(
    "eval_metrics, ordered_metric_names, expected_metric_names, expected_columns",
    [
        ({"src1": {}}, None, [], ["src1"]),
        ({"src1": {}, "src2": {"m1": 1}}, None, ["m1"], ["src1", "src2"]),
        (
            {"src1": {"m1": 1, "m2": 2, "m3": 3}, "src2": {"m1": 10, "m3": 30, "m4": 40}},
            None,
            ["m1", "m2", "m3", "m4"],
            ["src1", "src2"],
        ),
        (
            {"src1": {"m1": 1, "m2": 2, "m3": 3}, "src2": {"m1": 10, "m3": 30, "m4": 40}},
            ["m4", "m3", "M100"],
            ["m4", "m3", "m1", "m2"],
            ["src1", "src2"],
        ),
    ],
)
def test_get_merged_eval_metrics_works(
    eval_metrics, ordered_metric_names, expected_metric_names, expected_columns
):
    merged_df = get_merged_eval_metrics(eval_metrics, ordered_metric_names=ordered_metric_names)
    assert list(merged_df.index) == expected_metric_names
    assert list(merged_df.columns) == expected_columns


@pytest.mark.parametrize(
    "data_frame, max_cells, max_cols, max_rows, expected_cols, expected_rows",
    [
        (DataFrame(np.arange(160).reshape(8, 20)), 1000, 1, 1, 1, 1),
        (DataFrame(np.arange(160).reshape(8, 20)), 1000, 5, 4, 5, 4),
        (DataFrame(np.arange(160).reshape(8, 20)), 1000, 20, 4, 20, 4),
        (DataFrame(np.arange(160).reshape(8, 20)), 1000, 5, 8, 5, 8),
        (DataFrame(np.arange(160).reshape(8, 20)), 1000, 20, 8, 20, 8),
        (DataFrame(np.arange(160).reshape(8, 20)), 1000, 21, 8, 20, 8),
        (DataFrame(np.arange(160).reshape(8, 20)), 1000, 20, 9, 20, 8),
        (DataFrame(np.arange(160).reshape(8, 20)), 10, 5, 9, 5, 2),
        (DataFrame(np.arange(160).reshape(8, 20)), 10, 5, 1, 5, 1),
        (DataFrame(np.arange(160).reshape(8, 20)), 10, 30, 1, 20, 1),
        (DataFrame(np.arange(160).reshape(2, 80)), 80, 8, 10, 8, 2),
    ],
)
def test_get_data_profile_truncates_large_data_frame(
    data_frame, max_cells, max_cols, max_rows, expected_cols, expected_rows
):
    step_utils._MAX_PROFILE_CELL_SIZE = max_cells
    step_utils._MAX_PROFILE_COL_SIZE = max_cols
    step_utils._MAX_PROFILE_ROW_SIZE = max_rows
    profile = get_pandas_data_profile(data_frame, "fake profile")
    assert json.loads(profile.to_json())["table"]["n_var"] == expected_cols
    assert json.loads(profile.to_json())["variables"]["0"]["count"] == expected_rows


def test_get_data_profile_works_for_empty_data_frame():
    profile = get_pandas_data_profile(DataFrame(), "fake profile")
    assert json.loads(profile.to_json())["table"]["n_var"] == 1  # The index
    assert json.loads(profile.to_json())["variables"]["df_index"]["count"] == 0



"""
The ``mlflow.catboost`` module provides an API for logging and loading CatBoost models.
"""

import mlflow.catboost
from newron.models import ModelSignature, ModelInputExample
from collections import namedtuple
import inspect
from mlflow.utils.environment import (
    _mlflow_conda_env,
    _validate_env_arguments,
    _process_pip_requirements,
    _process_conda_env,
    _CONDA_ENV_FILE_NAME,
    _REQUIREMENTS_FILE_NAME,
    _CONSTRAINTS_FILE_NAME,
    _PYTHON_ENV_FILE_NAME,
    _PythonEnv,
)
from mlflow.utils.requirements_utils import _get_pinned_requirement

_load_model = mlflow.catboost._load_model
_CatboostModelWrapper = mlflow.catboost._CatboostModelWrapper
_load_pyfunc = mlflow.catboost._load_pyfunc
_init_model = mlflow.catboost._init_model

FLAVOR_NAME = "catboost"
DEFAULT_AWAIT_MAX_SLEEP_SECONDS = 5 * 60

def get_default_pip_requirements():
    """
    :return: A list of default pip requirements for Models produced by Catboost.
             Calls to :func:`save_model()` and :func:`log_model()` produce a pip environment
             that, at minimum, contains these requirements.
    """
    return [_get_pinned_requirement("catboost")]


def get_default_conda_env():
    """
    :return: The default Conda environment for MLflow Models produced by calls to
             :func:`save_model()` and :func:`log_model()`.
    """
    return _mlflow_conda_env(additional_pip_deps=get_default_pip_requirements())

def save_model(
    cb_model,
    path,
    conda_env=None,
    code_paths=None,
    mlflow_model=None,
    signature: ModelSignature = None,
    input_example: ModelInputExample = None,
    pip_requirements=None,
    extra_pip_requirements=None,
    **kwargs,
):
    """
    Save a CatBoost model to a path on the local file system.
    :param cb_model: CatBoost model (an instance of `CatBoost`_, `CatBoostClassifier`_,
                    `CatBoostRanker`_, or `CatBoostRegressor`_) to be saved.
    :param path: Local path where the model is to be saved.
    :param conda_env: {{ conda_env }}
    :param code_paths: A list of local filesystem paths to Python file dependencies (or directories
                       containing file dependencies). These files are *prepended* to the system
                       path when the model is loaded.
    :param mlflow_model: :py:mod:`mlflow.models.Model` this flavor is being added to.
    :param signature: :py:class:`ModelSignature <mlflow.models.ModelSignature>`
                      describes model input and output :py:class:`Schema <mlflow.types.Schema>`.
                      The model signature can be :py:func:`inferred <mlflow.models.infer_signature>`
                      from datasets with valid model input (e.g. the training dataset with target
                      column omitted) and valid model output (e.g. model predictions generated on
                      the training dataset), for example:
                      .. code-block:: python
                        from mlflow.models.signature import infer_signature
                        train = df.drop_column("target_label")
                        predictions = ... # compute model predictions
                        signature = infer_signature(train, predictions)
    :param input_example: Input example provides one or several instances of valid
                          model input. The example can be used as a hint of what data to feed the
                          model. The given example will be converted to a Pandas DataFrame and then
                          serialized to json using the Pandas split-oriented format. Bytes are
                          base64-encoded.
    :param pip_requirements: {{ pip_requirements }}
    :param extra_pip_requirements: {{ extra_pip_requirements }}
    :param kwargs: kwargs to pass to `CatBoost.save_model`_ method.
    """
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    del values["frame"]
    del values["kwargs"]
    mlflow.catboost.save_model(**values)

def log_model(
    cb_model,
    artifact_path,
    conda_env=None,
    code_paths=None,
    registered_model_name=None,
    signature: ModelSignature = None,
    input_example: ModelInputExample = None,
    await_registration_for=DEFAULT_AWAIT_MAX_SLEEP_SECONDS,
    pip_requirements=None,
    extra_pip_requirements=None,
    **kwargs,
):
    """
    Log a CatBoost model as an MLflow artifact for the current run.
    :param cb_model: CatBoost model (an instance of `CatBoost`_, `CatBoostClassifier`_,
                    `CatBoostRanker`_, or `CatBoostRegressor`_) to be saved.
    :param artifact_path: Run-relative artifact path.
    :param conda_env: {{ conda_env }}
    :param code_paths: A list of local filesystem paths to Python file dependencies (or directories
                       containing file dependencies). These files are *prepended* to the system
                       path when the model is loaded.
    :param registered_model_name: This argument may change or be removed in a
                                  future release without warning. If given, create a model
                                  version under ``registered_model_name``, also creating a
                                  registered model if one with the given name does not exist.
    :param signature: :py:class:`ModelSignature <mlflow.models.ModelSignature>`
                      describes model input and output :py:class:`Schema <mlflow.types.Schema>`.
                      The model signature can be :py:func:`inferred <mlflow.models.infer_signature>`
                      from datasets with valid model input (e.g. the training dataset with target
                      column omitted) and valid model output (e.g. model predictions generated on
                      the training dataset), for example:
                      .. code-block:: python
                        from mlflow.models.signature import infer_signature
                        train = df.drop_column("target_label")
                        predictions = ... # compute model predictions
                        signature = infer_signature(train, predictions)
    :param input_example: Input example provides one or several instances of valid
                          model input. The example can be used as a hint of what data to feed the
                          model. The given example will be converted to a Pandas DataFrame and then
                          serialized to json using the Pandas split-oriented format. Bytes are
                          base64-encoded.
    :param await_registration_for: Number of seconds to wait for the model version to finish
                        being created and is in ``READY`` status. By default, the function
                        waits for five minutes. Specify 0 or None to skip waiting.
    :param pip_requirements: {{ pip_requirements }}
    :param extra_pip_requirements: {{ extra_pip_requirements }}
    :param kwargs: kwargs to pass to `CatBoost.save_model`_ method.
    :return: A :py:class:`ModelInfo <mlflow.models.model.ModelInfo>` instance that contains the
             metadata of the logged model.
    """
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    del values["frame"]
    del values["kwargs"]
    return mlflow.catboost.log_model(**values)

def load_model(model_uri, dst_path=None):
    """
    Load a CatBoost model from a local file or a run.
    :param model_uri: The location, in URI format, of the MLflow model. For example:
                      - ``/Users/me/path/to/local/model``
                      - ``relative/path/to/local/model``
                      - ``s3://my_bucket/path/to/model``
                      - ``runs:/<mlflow_run_id>/run-relative/path/to/model``
                      For more information about supported URI schemes, see
                      `Referencing Artifacts <https://www.mlflow.org/docs/latest/tracking.html#
                      artifact-locations>`_.
    :param dst_path: The local filesystem path to which to download the model artifact.
                     This directory must already exist. If unspecified, a local output
                     path will be created.
    :return: A CatBoost model (an instance of `CatBoost`_, `CatBoostClassifier`_, `CatBoostRanker`_,
             or `CatBoostRegressor`_)
    """
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    del values["frame"]
    return mlflow.catboost.load_model(**values)


import mlflow.gluon

load_model = mlflow.gluon.load_model
_GluonModelWrapper = mlflow.gluon._GluonModelWrapper
_load_pyfunc = mlflow.gluon._load_pyfunc
save_model = mlflow.gluon.save_model
get_default_pip_requirements = mlflow.gluon.get_default_pip_requirements
get_default_conda_env = mlflow.gluon.get_default_conda_env
log_model = mlflow.gluon.log_model
autolog = mlflow.gluon.autolog




import os
import pathlib
from distutils.dir_util import copy_tree

import pandas as pd
import pytest
import yaml
from unittest import mock

import mlflow
from mlflow.entities import Run, SourceType
from mlflow.entities.model_registry import ModelVersion
from mlflow.exceptions import MlflowException
from mlflow.pipelines.pipeline import Pipeline
from mlflow.pipelines.step import BaseStep
from mlflow.pipelines.utils.execution import get_step_output_path, _get_execution_directory_basename
from mlflow.tracking.client import MlflowClient
from mlflow.tracking.context.registry import resolve_tags
from mlflow.utils.file_utils import path_to_local_file_uri
from mlflow.utils.mlflow_tags import (
    MLFLOW_SOURCE_NAME,
    MLFLOW_GIT_COMMIT,
    MLFLOW_GIT_REPO_URL,
    LEGACY_MLFLOW_GIT_REPO_URL,
    MLFLOW_SOURCE_TYPE,
)

# pylint: disable=unused-import
from tests.pipelines.helper_functions import (
    enter_pipeline_example_directory,
    enter_test_pipeline_directory,
    list_all_artifacts,
    chdir,
)  # pylint: enable=unused-import

_STEP_NAMES = ["ingest", "split", "train", "transform", "evaluate"]


@pytest.mark.usefixtures("enter_pipeline_example_directory")
def test_create_pipeline_fails_with_invalid_profile():
    with pytest.raises(
        MlflowException,
        match=r"(Failed to find|Did not find the YAML configuration)",
    ):
        Pipeline(profile="local123")


@pytest.mark.usefixtures("enter_pipeline_example_directory")
def test_create_pipeline_and_clean_works():
    p = Pipeline(profile="local")
    p.clean()


@pytest.mark.usefixtures("enter_pipeline_example_directory")
@pytest.mark.parametrize("empty_profile", [None, ""])
def test_create_pipeline_fails_with_empty_profile_name(empty_profile):
    with pytest.raises(MlflowException, match="A profile name must be provided"):
        _ = Pipeline(profile=empty_profile)


@pytest.mark.usefixtures("enter_pipeline_example_directory")
def test_create_pipeline_fails_with_path_containing_space(tmp_path):
    space_parent = tmp_path / "space parent"
    space_path = space_parent / "child"
    os.makedirs(space_parent, exist_ok=True)
    os.makedirs(space_path, exist_ok=True)
    copy_tree(os.getcwd(), str(space_path))

    with chdir(space_path), pytest.raises(
        MlflowException, match="Pipeline directory path cannot contain spaces"
    ):
        Pipeline(profile="local")


@pytest.mark.usefixtures("enter_pipeline_example_directory")
@pytest.mark.parametrize("custom_execution_directory", [None, "custom"])
def test_pipelines_execution_directory_is_managed_as_expected(
    custom_execution_directory, enter_pipeline_example_directory, tmp_path
):
    if custom_execution_directory is not None:
        custom_execution_directory = tmp_path / custom_execution_directory

    if custom_execution_directory is not None:
        os.environ["MLFLOW_PIPELINES_EXECUTION_DIRECTORY"] = str(custom_execution_directory)

    expected_execution_directory_location = (
        pathlib.Path(custom_execution_directory)
        if custom_execution_directory
        else pathlib.Path.home()
        / ".mlflow"
        / "pipelines"
        / _get_execution_directory_basename(enter_pipeline_example_directory)
    )

    # Run the full pipeline and verify that outputs for each step were written to the expected
    # execution directory locations
    p = Pipeline(profile="local")
    p.run()
    assert (expected_execution_directory_location / "Makefile").exists()
    assert (expected_execution_directory_location / "steps").exists()
    for step_name in _STEP_NAMES:
        step_outputs_path = expected_execution_directory_location / "steps" / step_name / "outputs"
        assert step_outputs_path.exists()
        first_output = next(step_outputs_path.iterdir(), None)
        # TODO: Assert that the ingest step has outputs once ingest execution has been implemented
        assert first_output is not None or step_name == "ingest"

    # Clean the pipeline and verify that all step outputs have been removed
    p.clean()
    for step_name in _STEP_NAMES:
        step_outputs_path = expected_execution_directory_location / "steps" / step_name / "outputs"
        assert not list(step_outputs_path.iterdir())


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_pipelines_log_to_expected_mlflow_backend_with_expected_run_tags_once_on_reruns(
    tmp_path,
):
    experiment_name = "my_test_exp"
    tracking_uri = "sqlite:///" + str((tmp_path / "tracking_dst.db").resolve())
    artifact_location = str((tmp_path / "mlartifacts").resolve())

    profile_path = pathlib.Path.cwd() / "profiles" / "local.yaml"
    with open(profile_path, "r") as f:
        profile_contents = yaml.safe_load(f)

    profile_contents["experiment"]["name"] = experiment_name
    profile_contents["experiment"]["tracking_uri"] = tracking_uri
    profile_contents["experiment"]["artifact_location"] = path_to_local_file_uri(artifact_location)

    with open(profile_path, "w") as f:
        yaml.safe_dump(profile_contents, f)

    mlflow.set_tracking_uri(tracking_uri)
    pipeline = Pipeline(profile="local")
    pipeline.clean()
    pipeline.run()

    logged_runs = mlflow.search_runs(experiment_names=[experiment_name], output_format="list")
    assert len(logged_runs) == 1
    logged_run = logged_runs[0]
    assert logged_run.info.artifact_uri == path_to_local_file_uri(
        str((pathlib.Path(artifact_location) / logged_run.info.run_id / "artifacts").resolve())
    )
    assert "r2_score_on_data_test" in logged_run.data.metrics
    artifacts = MlflowClient(tracking_uri).list_artifacts(
        run_id=logged_run.info.run_id, path="train"
    )
    assert {artifact.path for artifact in artifacts} == {
        "train/card.html",
        "train/estimator",
        "train/model",
    }
    run_tags = MlflowClient(tracking_uri).get_run(run_id=logged_run.info.run_id).data.tags
    pipelineSourceTag = {MLFLOW_SOURCE_TYPE: SourceType.to_string(SourceType.PIPELINE)}
    assert resolve_tags(pipelineSourceTag).items() <= run_tags.items()

    pipeline.run()
    logged_runs = mlflow.search_runs(experiment_names=[experiment_name], output_format="list")
    assert len(logged_runs) == 1


@pytest.mark.usefixtures("enter_pipeline_example_directory")
def test_pipelines_run_sets_mlflow_git_tags():
    pipeline = Pipeline(profile="local")
    pipeline.clean()
    pipeline.run(step="train")

    profile_path = pathlib.Path.cwd() / "profiles" / "local.yaml"
    with open(profile_path, "r") as f:
        profile_contents = yaml.safe_load(f)

    tracking_uri = profile_contents["experiment"]["tracking_uri"]
    experiment_name = profile_contents["experiment"]["name"]

    mlflow.set_tracking_uri(tracking_uri)
    logged_runs = mlflow.search_runs(
        experiment_names=[experiment_name],
        output_format="list",
        order_by=["attributes.start_time DESC"],
    )
    logged_run = logged_runs[0]

    run_tags = MlflowClient(tracking_uri).get_run(run_id=logged_run.info.run_id).data.tags
    assert {
        MLFLOW_SOURCE_NAME,
        MLFLOW_GIT_COMMIT,
        MLFLOW_GIT_REPO_URL,
        LEGACY_MLFLOW_GIT_REPO_URL,
    } < run_tags.keys()
    assert run_tags[MLFLOW_SOURCE_NAME] == run_tags[MLFLOW_GIT_REPO_URL]


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_pipelines_run_throws_exception_and_produces_failure_card_when_step_fails():
    profile_path = pathlib.Path.cwd() / "profiles" / "local.yaml"
    with open(profile_path, "r") as f:
        profile_contents = yaml.safe_load(f)

    profile_contents["INGEST_DATA_LOCATION"] = "a bad location"

    with open(profile_path, "w") as f:
        yaml.safe_dump(profile_contents, f)

    pipeline = Pipeline(profile="local")
    pipeline.clean()
    with pytest.raises(MlflowException, match="Failed to run.*test_pipeline.*ingest"):
        pipeline.run()
    with pytest.raises(MlflowException, match="Failed to run.*split.*test_pipeline.*ingest"):
        pipeline.run(step="split")

    step_card_path = get_step_output_path(
        pipeline_root_path=pipeline._pipeline_root_path,
        step_name="ingest",
        relative_path="card.html",
    )
    with open(step_card_path, "r") as f:
        card_content = f.read()

    assert "Ingest" in card_content
    assert "Failed" in card_content
    assert "Stacktrace" in card_content


@pytest.mark.usefixtures("enter_pipeline_example_directory")
def test_test_step_logs_step_cards_as_artifacts():
    pipeline = Pipeline(profile="local")
    pipeline.clean()
    pipeline.run()

    tracking_uri = pipeline._get_step("train").tracking_config.tracking_uri
    local_run_id_path = get_step_output_path(
        pipeline_root_path=pipeline._pipeline_root_path,
        step_name="train",
        relative_path="run_id",
    )
    run_id = pathlib.Path(local_run_id_path).read_text()
    artifacts = set(list_all_artifacts(tracking_uri, run_id))
    assert artifacts.issuperset(
        {
            "ingest/card.html",
            "split/card.html",
            "transform/card.html",
            "train/card.html",
            "evaluate/card.html",
            "register/card.html",
        }
    )


@pytest.mark.usefixtures("enter_pipeline_example_directory")
def test_pipeline_get_artifacts():
    pipeline = Pipeline(profile="local")
    pipeline.clean()

    pipeline.run("ingest")
    pipeline.run("split")
    pipeline.run("transform")
    pipeline.run("train")
    pipeline.run("register")

    assert isinstance(pipeline.get_artifact("ingested_data"), pd.DataFrame)
    assert isinstance(pipeline.get_artifact("training_data"), pd.DataFrame)
    assert isinstance(pipeline.get_artifact("validation_data"), pd.DataFrame)
    assert isinstance(pipeline.get_artifact("test_data"), pd.DataFrame)
    assert isinstance(pipeline.get_artifact("transformed_training_data"), pd.DataFrame)
    assert isinstance(pipeline.get_artifact("transformed_validation_data"), pd.DataFrame)
    assert hasattr(pipeline.get_artifact("transformer"), "transform")
    assert isinstance(pipeline.get_artifact("model"), mlflow.pyfunc.PyFuncModel)
    assert isinstance(pipeline.get_artifact("run"), Run)
    assert isinstance(pipeline.get_artifact("registered_model_version"), ModelVersion)

    with pytest.raises(MlflowException, match="The artifact with name 'abcde' is not supported."):
        pipeline.get_artifact("abcde")

    pipeline.clean()
    with mock.patch("mlflow.pipelines.regression.v1.pipeline._logger.warning") as mock_warning:
        pipeline.get_artifact("ingested_data")
        mock_warning.assert_called_once_with(
            "The artifact with name 'ingested_data' was not found."
            " Re-run the 'ingest' step to generate it."
        )


def test_generate_worst_examples_dataframe():
    test_df = pd.DataFrame(
        {
            "a": [3, 2, 5],
            "b": [6, 9, 1],
        }
    )
    target_col = "b"
    predictions = [5, 3, 4]

    result_df = BaseStep._generate_worst_examples_dataframe(
        test_df, predictions, target_col, worst_k=2
    )

    def assert_result_correct(df):
        assert df.columns.tolist() == ["absolute_error", "prediction", "b", "a"]
        assert df.absolute_error.tolist() == [6, 3]
        assert df.prediction.tolist() == [3, 4]
        assert df.b.tolist() == [9, 1]
        assert df.a.tolist() == [2, 5]
        assert df.index.tolist() == [0, 1]

    assert_result_correct(result_df)

    test_df2 = test_df.set_axis([2, 1, 0], axis="index")
    result_df2 = BaseStep._generate_worst_examples_dataframe(
        test_df2, predictions, target_col, worst_k=2
    )
    assert_result_correct(result_df2)



import os
import random
import functools
from unittest import mock
from contextlib import ExitStack, contextmanager


import logging
import requests
import time
import signal
import socket
import subprocess
import uuid
import sys
import yaml

import pytest

import newron
import mlflow
from mlflow.tracking.artifact_utils import _download_artifact_from_uri
from mlflow.utils.file_utils import read_yaml, write_yaml
from mlflow.utils.environment import (
    _get_pip_deps,
    _CONDA_ENV_FILE_NAME,
    _REQUIREMENTS_FILE_NAME,
    _CONSTRAINTS_FILE_NAME,
)

AWS_METADATA_IP = "169.254.169.254"  # Used to fetch AWS Instance and User metadata.
LOCALHOST = "127.0.0.1"
PROTOBUF_REQUIREMENT = "protobuf<4.0.0"

_logger = logging.getLogger(__name__)


def get_safe_port():
    """Returns an ephemeral port that is very likely to be free to bind to."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind((LOCALHOST, 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def random_int(lo=1, hi=1e10):
    return random.randint(lo, hi)


def random_str(size=10):
    msg = (
        "UUID4 generated strings have a high potential for collision at small sizes."
        "10 is set as the lower bounds for random string generation to prevent non-deterministic"
        "test failures."
    )
    assert size >= 10, msg
    return uuid.uuid4().hex[:size]


def random_file(ext):
    return "temp_test_%d.%s" % (random_int(), ext)


def score_model_in_sagemaker_docker_container(
    model_uri,
    data,
    content_type,
    flavor="python_function",
    activity_polling_timeout_seconds=500,
):
    """
    :param model_uri: URI to the model to be served.
    :param data: The data to send to the docker container for testing. This is either a
                 Pandas dataframe or string of the format specified by `content_type`.
    :param content_type: The type of the data to send to the docker container for testing. This is
                         one of `mlflow.pyfunc.scoring_server.CONTENT_TYPES`.
    :param flavor: Model flavor to be deployed.
    :param activity_polling_timeout_seconds: The amount of time, in seconds, to wait before
                                             declaring the scoring process to have failed.
    """
    env = dict(os.environ)
    env.update(LC_ALL="en_US.UTF-8", LANG="en_US.UTF-8")
    proc = _start_scoring_proc(
        cmd=["mlflow", "sagemaker", "run-local", "-m", model_uri, "-p", "5000", "-f", flavor],
        env=env,
    )
    return _evaluate_scoring_proc(proc, 5000, data, content_type, activity_polling_timeout_seconds)


def pyfunc_build_image(model_uri=None, extra_args=None):
    """
    Builds a docker image containing the specified model, returning the name of the image.
    :param model_uri: URI of model, e.g. runs:/some-run-id/run-relative/path/to/model
    :param extra_args: List of extra args to pass to `mlflow models build-docker` command
    """
    name = uuid.uuid4().hex
    cmd = [
        "mlflow",
        "models",
        "build-docker",
        *(["-m", model_uri] if model_uri else []),
        "-n",
        name,
    ]
    mlflow_home = os.environ.get("MLFLOW_HOME")
    if mlflow_home:
        cmd += ["--mlflow-home", mlflow_home]
    if extra_args:
        cmd += extra_args
    p = subprocess.Popen(cmd)
    assert p.wait() == 0, "Failed to build docker image to serve model from %s" % model_uri
    return name


def pyfunc_serve_from_docker_image(image_name, host_port, extra_args=None):
    """
    Serves a model from a docker container, exposing it as an endpoint at the specified port
    on the host machine. Returns a handle (Popen object) to the server process.
    """
    env = dict(os.environ)
    env.update(LC_ALL="en_US.UTF-8", LANG="en_US.UTF-8")
    scoring_cmd = ["docker", "run", "-p", "%s:8080" % host_port, image_name]
    if extra_args is not None:
        scoring_cmd += extra_args
    return _start_scoring_proc(cmd=scoring_cmd, env=env)


def pyfunc_serve_from_docker_image_with_env_override(
    image_name, host_port, gunicorn_opts, extra_args=None, extra_docker_run_options=None
):
    """
    Serves a model from a docker container, exposing it as an endpoint at the specified port
    on the host machine. Returns a handle (Popen object) to the server process.
    """
    env = dict(os.environ)
    env.update(LC_ALL="en_US.UTF-8", LANG="en_US.UTF-8")
    scoring_cmd = [
        "docker",
        "run",
        "-e",
        "GUNICORN_CMD_ARGS=%s" % gunicorn_opts,
        "-p",
        "%s:8080" % host_port,
        *(extra_docker_run_options or []),
        image_name,
    ]
    if extra_args is not None:
        scoring_cmd += extra_args
    return _start_scoring_proc(cmd=scoring_cmd, env=env)


def pyfunc_serve_and_score_model(
    model_uri,
    data,
    content_type,
    activity_polling_timeout_seconds=500,
    extra_args=None,
    stdout=sys.stdout,
):
    """
    :param model_uri: URI to the model to be served.
    :param data: The data to send to the pyfunc server for testing. This is either a
                 Pandas dataframe or string of the format specified by `content_type`.
    :param content_type: The type of the data to send to the pyfunc server for testing. This is
                         one of `mlflow.pyfunc.scoring_server.CONTENT_TYPES`.
    :param activity_polling_timeout_seconds: The amount of time, in seconds, to wait before
                                             declaring the scoring process to have failed.
    :param extra_args: A list of extra arguments to pass to the pyfunc scoring server command. For
                       example, passing ``extra_args=["--env-manager", "local"]`` will pass the
                       ``--env-manager local`` flag to the scoring server to ensure that conda
                       environment activation is skipped.
    """
    env = dict(os.environ)
    env.update(LC_ALL="en_US.UTF-8", LANG="en_US.UTF-8")
    env.update(MLFLOW_TRACKING_URI=mlflow.get_tracking_uri())
    env.update(MLFLOW_HOME=_get_mlflow_home())
    port = get_safe_port()
    scoring_cmd = [
        "mlflow",
        "models",
        "serve",
        "-m",
        model_uri,
        "-p",
        str(port),
        "--install-mlflow",
    ]
    if extra_args is not None:
        scoring_cmd += extra_args
    proc = _start_scoring_proc(cmd=scoring_cmd, env=env, stdout=stdout, stderr=stdout)
    return _evaluate_scoring_proc(proc, port, data, content_type, activity_polling_timeout_seconds)


def _get_mlflow_home():
    """
    :return: The path to the MLflow installation root directory
    """
    mlflow_module_path = os.path.dirname(os.path.abspath(mlflow.__file__))
    # The MLflow root directory is one level about the mlflow module location
    return os.path.join(mlflow_module_path, os.pardir)


def _start_scoring_proc(cmd, env, stdout=sys.stdout, stderr=sys.stderr):
    if os.name != "nt":
        return subprocess.Popen(
            cmd,
            stdout=stdout,
            stderr=stderr,
            text=True,
            env=env,
            # Assign the scoring process to a process group. All child processes of the
            # scoring process will be assigned to this group as well. This allows child
            # processes of the scoring process to be terminated successfully
            preexec_fn=os.setsid,
        )
    else:
        return subprocess.Popen(
            cmd,
            stdout=stdout,
            stderr=stderr,
            text=True,
            env=env,
            # On Windows, `os.setsid` and `preexec_fn` are unavailable
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
        )


class RestEndpoint:
    def __init__(self, proc, port, activity_polling_timeout_seconds=250):
        self._proc = proc
        self._port = port
        self._activity_polling_timeout_seconds = activity_polling_timeout_seconds

    def __enter__(self):
        for i in range(self._activity_polling_timeout_seconds):
            assert self._proc.poll() is None, "scoring process died"
            time.sleep(1)
            # noinspection PyBroadException
            try:
                ping_status = requests.get(url="http://localhost:%d/ping" % self._port)
                _logger.info(f"connection attempt {i} server is up! ping status {ping_status}")
                if ping_status.status_code == 200:
                    break
            except Exception:
                _logger.info(f"connection attempt {i} failed, server is not up yet")
        if ping_status.status_code != 200:
            raise Exception("ping failed, server is not happy")
        _logger.info(f"server up, ping status {ping_status}")
        return self

    def __exit__(self, tp, val, traceback):
        if self._proc.poll() is None:
            # Terminate the process group containing the scoring process.
            # This will terminate all child processes of the scoring process
            if os.name != "nt":
                pgrp = os.getpgid(self._proc.pid)
                os.killpg(pgrp, signal.SIGTERM)
            else:
                # https://stackoverflow.com/questions/47016723/windows-equivalent-for-spawning-and-killing-separate-process-group-in-python-3
                self._proc.send_signal(signal.CTRL_BREAK_EVENT)
                self._proc.kill()

    def invoke(self, data, content_type):
        import mlflow.pyfunc.scoring_server as pyfunc_scoring_server
        import pandas as pd

        if type(data) == pd.DataFrame:
            if content_type == pyfunc_scoring_server.CONTENT_TYPE_JSON_RECORDS_ORIENTED:
                data = data.to_json(orient="records")
            elif (
                content_type == pyfunc_scoring_server.CONTENT_TYPE_JSON
                or content_type == pyfunc_scoring_server.CONTENT_TYPE_JSON_SPLIT_ORIENTED
            ):
                data = data.to_json(orient="split")
            elif content_type == pyfunc_scoring_server.CONTENT_TYPE_CSV:
                data = data.to_csv(index=False)
            else:
                raise Exception(
                    "Unexpected content type for Pandas dataframe input %s" % content_type
                )
        response = requests.post(
            url="http://localhost:%d/invocations" % self._port,
            data=data,
            headers={"Content-Type": content_type},
        )
        return response


def _evaluate_scoring_proc(proc, port, data, content_type, activity_polling_timeout_seconds=250):
    """
    :param activity_polling_timeout_seconds: The amount of time, in seconds, to wait before
                                             declaring the scoring process to have failed.
    """
    with RestEndpoint(proc, port, activity_polling_timeout_seconds) as endpoint:
        return endpoint.invoke(data, content_type)


@pytest.fixture(scope="function", autouse=True)
def set_boto_credentials(monkeypatch):
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "NotARealAccessKey")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "NotARealSecretAccessKey")
    monkeypatch.setenv("AWS_SESSION_TOKEN", "NotARealSessionToken")


class safe_edit_yaml:
    def __init__(self, root, file_name, edit_func):
        self._root = root
        self._file_name = file_name
        self._edit_func = edit_func
        self._original = read_yaml(root, file_name)

    def __enter__(self):
        new_dict = self._edit_func(self._original.copy())
        write_yaml(self._root, self._file_name, new_dict, overwrite=True)

    def __exit__(self, *args):
        write_yaml(self._root, self._file_name, self._original, overwrite=True)


def create_mock_response(status_code, text):
    """
    Create a mock response object with the status_code and text

    :param: status_code int HTTP status code
    :param: text message from the response
    :return: mock HTTP Response
    """
    response = mock.MagicMock()
    response.status_code = status_code
    response.text = text
    return response


def _read_yaml(path):
    with open(path, "r") as f:
        return yaml.safe_load(f)


def _read_lines(path):
    with open(path, "r") as f:
        return f.read().splitlines()


def _compare_logged_code_paths(code_path, model_path, flavor_name):
    import mlflow.pyfunc
    from mlflow.utils.model_utils import _get_flavor_configuration, FLAVOR_CONFIG_CODE

    pyfunc_conf = _get_flavor_configuration(
        model_path=model_path, flavor_name=mlflow.pyfunc.FLAVOR_NAME
    )
    flavor_conf = _get_flavor_configuration(model_path, flavor_name=flavor_name)
    assert pyfunc_conf[mlflow.pyfunc.CODE] == flavor_conf[FLAVOR_CONFIG_CODE]
    saved_code_path = os.path.join(model_path, pyfunc_conf[mlflow.pyfunc.CODE])
    assert os.path.exists(saved_code_path)

    with open(os.path.join(saved_code_path, os.path.basename(code_path)), "r") as f1:
        with open(code_path, "r") as f2:
            assert f1.read() == f2.read()


def _compare_conda_env_requirements(env_path, req_path):
    assert os.path.exists(req_path)
    custom_env_parsed = _read_yaml(env_path)
    requirements = _read_lines(req_path)
    assert _get_pip_deps(custom_env_parsed) == requirements


def _assert_pip_requirements(model_uri, requirements, constraints=None, strict=False):
    """
    Loads the pip requirements (and optionally constraints) from `model_uri` and compares them
    to `requirements` (and `constraints`).

    If `strict` is True, evaluate `set(requirements) == set(loaded_requirements)`.
    Otherwise, evaluate `set(requirements) <= set(loaded_requirements)`.
    """
    local_path = _download_artifact_from_uri(model_uri)
    txt_reqs = _read_lines(os.path.join(local_path, _REQUIREMENTS_FILE_NAME))
    conda_reqs = _get_pip_deps(_read_yaml(os.path.join(local_path, _CONDA_ENV_FILE_NAME)))
    compare_func = set.__eq__ if strict else set.__le__
    requirements = set(requirements)
    assert compare_func(requirements, set(txt_reqs))
    assert compare_func(requirements, set(conda_reqs))

    if constraints is not None:
        assert f"-c {_CONSTRAINTS_FILE_NAME}" in txt_reqs
        assert f"-c {_CONSTRAINTS_FILE_NAME}" in conda_reqs
        cons = _read_lines(os.path.join(local_path, _CONSTRAINTS_FILE_NAME))
        assert compare_func(set(constraints), set(cons))


def _is_available_on_pypi(package, version=None, module=None):
    """
    Returns True if the specified package version is available on PyPI.

    :param package: The name of the package.
    :param version: The version of the package. If None, defaults to the installed version.
    :param module: The name of the top-level module provided by the package . For example,
                   if `package` is 'scikit-learn', `module` should be 'sklearn'. If None, defaults
                   to `package`.
    """
    from mlflow.utils.requirements_utils import _get_installed_version

    resp = requests.get("https://pypi.python.org/pypi/{}/json".format(package))
    if not resp.ok:
        return False

    version = version or _get_installed_version(module or package)
    dist_files = resp.json()["releases"].get(version)
    return (
        dist_files is not None  # specified version exists
        and (len(dist_files) > 0)  # at least one distribution file exists
        and not dist_files[0].get("yanked", False)  # specified version is not yanked
    )


def _is_importable(module_name):
    try:
        __import__(module_name)
        return True
    except ImportError:
        return False


def allow_infer_pip_requirements_fallback_if(condition):
    def decorator(f):
        return pytest.mark.allow_infer_pip_requirements_fallback(f) if condition else f

    return decorator


def mock_method_chain(mock_obj, methods, return_value=None, side_effect=None):
    """
    Mock a chain of methods.

    Examples
    --------
    >>> from unittest import mock
    >>> m = mock.MagicMock()
    >>> mock_method_chain(m, ["a", "b"], return_value=0)
    >>> m.a().b()
    0
    >>> mock_method_chain(m, ["c.d", "e"], return_value=1)
    >>> m.c.d().e()
    1
    >>> mock_method_chain(m, ["f"], side_effect=Exception("side_effect"))
    >>> m.f()
    Traceback (most recent call last):
      ...
    Exception: side_effect
    """
    length = len(methods)
    for idx, method in enumerate(methods):
        mock_obj = functools.reduce(getattr, method.split("."), mock_obj)
        if idx != length - 1:
            mock_obj = mock_obj.return_value
        else:
            mock_obj.return_value = return_value
            mock_obj.side_effect = side_effect


@contextmanager
def multi_context(*cms):
    with ExitStack() as stack:
        yield list(map(stack.enter_context, cms))


class StartsWithMatcher:
    def __init__(self, prefix):
        self.prefix = prefix

    def __eq__(self, other):
        return isinstance(other, str) and other.startswith(self.prefix)


class AnyStringWith(str):
    def __eq__(self, other):
        return self in other


from collections import namedtuple
from unittest import mock
from packaging.version import Version
import os
import pytest
import yaml

import catboost as cb
import numpy as np
import pandas as pd
from sklearn import datasets
from sklearn.pipeline import Pipeline

import newron.catboost
from newron import pyfunc
import mlflow.pyfunc.scoring_server as pyfunc_scoring_server
from mlflow.models.utils import _read_example
from newron.models import Model, infer_signature
from mlflow.store.artifact.s3_artifact_repo import S3ArtifactRepository
from mlflow.tracking.artifact_utils import _download_artifact_from_uri
from mlflow.utils.environment import _mlflow_conda_env
from mlflow.utils.model_utils import _get_flavor_configuration
from mlflow.tracking._model_registry import DEFAULT_AWAIT_MAX_SLEEP_SECONDS

from tests.helper_functions import (
    pyfunc_serve_and_score_model,
    _compare_conda_env_requirements,
    _assert_pip_requirements,
    _is_available_on_pypi,
    _compare_logged_code_paths,
)

EXTRA_PYFUNC_SERVING_TEST_ARGS = (
    [] if _is_available_on_pypi("catboost") else ["--env-manager", "local"]
)

ModelWithData = namedtuple("ModelWithData", ["model", "inference_dataframe"])


def get_iris():
    iris = datasets.load_iris()
    X = pd.DataFrame(iris.data[:, :2], columns=iris.feature_names[:2])
    y = pd.Series(iris.target)
    return X, y


def read_yaml(path):
    with open(path, "r") as f:
        return yaml.safe_load(f)


MODEL_PARAMS = {"allow_writing_files": False, "iterations": 10}


@pytest.fixture(
    scope="module",
    params=[
        cb.CatBoost(MODEL_PARAMS),
        cb.CatBoostClassifier(**MODEL_PARAMS),
        cb.CatBoostRegressor(**MODEL_PARAMS),
    ],
    ids=["CatBoost", "CatBoostClassifier", "CatBoostRegressor"],
)
def cb_model(request):
    model = request.param
    X, y = get_iris()
    return ModelWithData(model=model.fit(X, y), inference_dataframe=X)


@pytest.fixture
def reg_model():
    model = cb.CatBoostRegressor(**MODEL_PARAMS)
    X, y = get_iris()
    return ModelWithData(model=model.fit(X, y), inference_dataframe=X)


@pytest.fixture
def model_path(tmpdir):
    return tmpdir.join("model").strpath


@pytest.fixture
def custom_env(tmpdir):
    conda_env_path = os.path.join(tmpdir.strpath, "conda_env.yml")
    _mlflow_conda_env(conda_env_path, additional_pip_deps=["catboost", "pytest"])
    return conda_env_path


@pytest.mark.parametrize("model_type", ["CatBoost", "CatBoostClassifier", "CatBoostRegressor"])
def test_init_model(model_type):
    model = newron.catboost._init_model(model_type)
    assert model.__class__.__name__ == model_type


@pytest.mark.skipif(
    Version(cb.__version__) < Version("0.26.0"),
    reason="catboost < 0.26.0 does not support CatBoostRanker",
)
def test_log_catboost_ranker():
    """
    This is a separate test for the CatBoostRanker model.
    It is separate since the ranking task requires a group_id column which makes the code different.
    """
    # the ranking task requires setting a group_id
    # we are creating a dummy group_id here that doesn't make any sense for the Iris dataset,
    # but is ok for testing if the code is running correctly
    X, y = get_iris()
    dummy_group_id = np.arange(len(X)) % 3
    dummy_group_id.sort()

    model = cb.CatBoostRanker(**MODEL_PARAMS, subsample=1.0)
    model.fit(X, y, group_id=dummy_group_id)

    with newron.start_run():
        model_info = newron.catboost.log_model(model, "model")
        loaded_model = newron.catboost.load_model(model_info.model_uri)
        assert isinstance(loaded_model, cb.CatBoostRanker)
        np.testing.assert_array_almost_equal(model.predict(X), loaded_model.predict(X))


def test_init_model_throws_for_invalid_model_type():
    with pytest.raises(TypeError, match="Invalid model type"):
        newron.catboost._init_model("unsupported")


def test_model_save_load(cb_model, model_path):
    model, inference_dataframe = cb_model
    newron.catboost.save_model(cb_model=model, path=model_path)

    loaded_model = newron.catboost.load_model(model_uri=model_path)
    np.testing.assert_array_almost_equal(
        model.predict(inference_dataframe),
        loaded_model.predict(inference_dataframe),
    )

    loaded_pyfunc = pyfunc.load_model(model_uri=model_path)
    np.testing.assert_array_almost_equal(
        loaded_model.predict(inference_dataframe),
        loaded_pyfunc.predict(inference_dataframe),
    )


def test_log_model_logs_model_type(cb_model):
    with newron.start_run():
        artifact_path = "model"
        newron.catboost.log_model(cb_model.model, artifact_path)
        model_uri = newron.get_artifact_uri(artifact_path)

    flavor_conf = Model.load(model_uri).flavors["catboost"]
    assert "model_type" in flavor_conf
    assert flavor_conf["model_type"] == cb_model.model.__class__.__name__


# Supported serialization formats:
# https://catboost.ai/docs/concepts/python-reference_catboost_save_model.html
SUPPORTS_DESERIALIZATION = ["cbm", "coreml", "json", "onnx"]
save_formats = SUPPORTS_DESERIALIZATION + ["python", "cpp", "pmml"]


@pytest.mark.allow_infer_pip_requirements_fallback
@pytest.mark.parametrize("save_format", save_formats)
def test_log_model_logs_save_format(reg_model, save_format):
    with newron.start_run():
        artifact_path = "model"
        newron.catboost.log_model(reg_model.model, artifact_path, format=save_format)
        model_uri = newron.get_artifact_uri(artifact_path)

    flavor_conf = Model.load(model_uri).flavors["catboost"]
    assert "save_format" in flavor_conf
    assert flavor_conf["save_format"] == save_format

    if save_format in SUPPORTS_DESERIALIZATION:
        newron.catboost.load_model(model_uri)
    else:
        with pytest.raises(cb.CatBoostError, match="deserialization not supported or missing"):
            newron.catboost.load_model(model_uri)


@pytest.mark.parametrize("signature", [None, infer_signature(get_iris()[0])])
@pytest.mark.parametrize("input_example", [None, get_iris()[0].head(3)])
def test_signature_and_examples_are_saved_correctly(
    reg_model, model_path, signature, input_example
):
    newron.catboost.save_model(
        reg_model.model, model_path, signature=signature, input_example=input_example
    )
    newron_model = Model.load(model_path)
    assert signature == newron_model.signature
    if input_example is None:
        assert newron_model.saved_input_example_info is None
    else:
        pd.testing.assert_frame_equal(_read_example(newron_model, model_path), input_example)


def test_model_load_from_remote_uri_succeeds(reg_model, model_path, mock_s3_bucket):
    model, inference_dataframe = reg_model
    newron.catboost.save_model(cb_model=model, path=model_path)
    artifact_root = "s3://{bucket_name}".format(bucket_name=mock_s3_bucket)
    artifact_repo = S3ArtifactRepository(artifact_root)
    artifact_path = "model"
    artifact_repo.log_artifacts(model_path, artifact_path=artifact_path)

    model_uri = artifact_root + "/" + artifact_path
    loaded_model = newron.catboost.load_model(model_uri=model_uri)
    np.testing.assert_array_almost_equal(
        model.predict(inference_dataframe),
        loaded_model.predict(inference_dataframe),
    )


def test_log_model(cb_model, tmpdir):
    model, inference_dataframe = cb_model
    with newron.start_run():
        artifact_path = "model"
        conda_env = os.path.join(tmpdir.strpath, "conda_env.yaml")
        _mlflow_conda_env(conda_env, additional_pip_deps=["catboost"])

        model_info = newron.catboost.log_model(model, artifact_path, conda_env=conda_env)
        model_uri = "runs:/{}/{}".format(newron.active_run().info.run_id, artifact_path)
        assert model_info.model_uri == model_uri

        loaded_model = newron.catboost.load_model(model_uri)
        np.testing.assert_array_almost_equal(
            model.predict(inference_dataframe),
            loaded_model.predict(inference_dataframe),
        )

        local_path = _download_artifact_from_uri(model_uri)
        model_config = Model.load(os.path.join(local_path, "MLmodel"))
        assert pyfunc.FLAVOR_NAME in model_config.flavors
        assert pyfunc.ENV in model_config.flavors[pyfunc.FLAVOR_NAME]
        env_path = model_config.flavors[pyfunc.FLAVOR_NAME][pyfunc.ENV]
        assert os.path.exists(os.path.join(local_path, env_path))


def test_log_model_calls_register_model(cb_model, tmpdir):
    artifact_path = "model"
    registered_model_name = "registered_model"
    with newron.start_run() as run, mock.patch("newron.register_model") as register_model_mock:
        conda_env_path = os.path.join(tmpdir.strpath, "conda_env.yaml")
        _mlflow_conda_env(conda_env_path, additional_pip_deps=["catboost"])
        newron.catboost.log_model(
            cb_model.model,
            artifact_path,
            conda_env=conda_env_path,
            registered_model_name=registered_model_name,
        )
        model_uri = "runs:/{}/{}".format(run.info.run_id, artifact_path)
        register_model_mock.assert_called_once_with(
            model_uri, registered_model_name, await_registration_for=DEFAULT_AWAIT_MAX_SLEEP_SECONDS
        )


def test_log_model_no_registered_model_name(cb_model, tmpdir):
    with newron.start_run(), mock.patch("newron.register_model") as register_model_mock:
        artifact_path = "model"
        conda_env_path = os.path.join(tmpdir.strpath, "conda_env.yaml")
        _mlflow_conda_env(conda_env_path, additional_pip_deps=["catboost"])
        newron.catboost.log_model(cb_model.model, artifact_path, conda_env=conda_env_path)
        register_model_mock.assert_not_called()


def test_model_save_persists_specified_conda_env_in_newron_model_directory(
    reg_model, model_path, custom_env
):
    newron.catboost.save_model(cb_model=reg_model.model, path=model_path, conda_env=custom_env)
    pyfunc_conf = _get_flavor_configuration(model_path=model_path, flavor_name=pyfunc.FLAVOR_NAME)
    saved_conda_env_path = os.path.join(model_path, pyfunc_conf[pyfunc.ENV])
    assert os.path.exists(saved_conda_env_path)
    assert saved_conda_env_path != custom_env
    assert read_yaml(saved_conda_env_path) == read_yaml(custom_env)


def test_model_save_persists_requirements_in_newron_model_directory(
    reg_model, model_path, custom_env
):
    newron.catboost.save_model(cb_model=reg_model.model, path=model_path, conda_env=custom_env)

    saved_pip_req_path = os.path.join(model_path, "requirements.txt")
    _compare_conda_env_requirements(custom_env, saved_pip_req_path)


def test_model_save_accepts_conda_env_as_dict(reg_model, model_path):
    conda_env = newron.catboost.get_default_conda_env()
    conda_env["dependencies"].append("pytest")
    newron.catboost.save_model(cb_model=reg_model.model, path=model_path, conda_env=conda_env)

    pyfunc_conf = _get_flavor_configuration(model_path=model_path, flavor_name=pyfunc.FLAVOR_NAME)
    saved_conda_env_path = os.path.join(model_path, pyfunc_conf[pyfunc.ENV])
    assert os.path.exists(saved_conda_env_path)
    assert read_yaml(saved_conda_env_path) == conda_env


def test_model_log_persists_specified_conda_env_in_newron_model_directory(reg_model, custom_env):
    artifact_path = "model"
    with newron.start_run():
        newron.catboost.log_model(reg_model.model, artifact_path, conda_env=custom_env)
        model_uri = newron.get_artifact_uri(artifact_path)

    local_path = _download_artifact_from_uri(artifact_uri=model_uri)
    pyfunc_conf = _get_flavor_configuration(model_path=local_path, flavor_name=pyfunc.FLAVOR_NAME)
    saved_conda_env_path = os.path.join(local_path, pyfunc_conf[pyfunc.ENV])
    assert os.path.exists(saved_conda_env_path)
    assert saved_conda_env_path != custom_env
    assert read_yaml(saved_conda_env_path) == read_yaml(custom_env)


def test_model_log_persists_requirements_in_newron_model_directory(reg_model, custom_env):
    artifact_path = "model"
    with newron.start_run():
        newron.catboost.log_model(reg_model.model, artifact_path, conda_env=custom_env)
        model_uri = newron.get_artifact_uri(artifact_path)

    local_path = _download_artifact_from_uri(artifact_uri=model_uri)
    saved_pip_req_path = os.path.join(local_path, "requirements.txt")
    _compare_conda_env_requirements(custom_env, saved_pip_req_path)


def test_log_model_with_pip_requirements(reg_model, tmpdir):
    # Path to a requirements file
    req_file = tmpdir.join("requirements.txt")
    req_file.write("a")
    with newron.start_run():
        newron.catboost.log_model(reg_model.model, "model", pip_requirements=req_file.strpath)
        _assert_pip_requirements(newron.get_artifact_uri("model"), ["newron", "a"], strict=True)

    # List of requirements
    with newron.start_run():
        newron.catboost.log_model(
            reg_model.model, "model", pip_requirements=[f"-r {req_file.strpath}", "b"]
        )
        _assert_pip_requirements(
            newron.get_artifact_uri("model"), ["newron", "a", "b"], strict=True
        )

    # Constraints file
    with newron.start_run():
        newron.catboost.log_model(
            reg_model.model, "model", pip_requirements=[f"-c {req_file.strpath}", "b"]
        )
        _assert_pip_requirements(
            newron.get_artifact_uri("model"),
            ["newron", "b", "-c constraints.txt"],
            ["a"],
            strict=True,
        )


def test_log_model_with_extra_pip_requirements(reg_model, tmpdir):
    default_reqs = newron.catboost.get_default_pip_requirements()

    # Path to a requirements file
    req_file = tmpdir.join("requirements.txt")
    req_file.write("a")
    with newron.start_run():
        newron.catboost.log_model(reg_model.model, "model", extra_pip_requirements=req_file.strpath)
        _assert_pip_requirements(newron.get_artifact_uri("model"), ["newron", *default_reqs, "a"])

    # List of requirements
    with newron.start_run():
        newron.catboost.log_model(
            reg_model.model, "model", extra_pip_requirements=[f"-r {req_file.strpath}", "b"]
        )
        _assert_pip_requirements(
            newron.get_artifact_uri("model"), ["newron", *default_reqs, "a", "b"]
        )

    # Constraints file
    with newron.start_run():
        newron.catboost.log_model(
            reg_model.model, "model", extra_pip_requirements=[f"-c {req_file.strpath}", "b"]
        )
        _assert_pip_requirements(
            newron.get_artifact_uri("model"),
            ["newron", *default_reqs, "b", "-c constraints.txt"],
            ["a"],
        )


def test_model_save_without_specified_conda_env_uses_default_env_with_expected_dependencies(
    reg_model, model_path
):
    newron.catboost.save_model(reg_model.model, model_path)
    _assert_pip_requirements(model_path, newron.catboost.get_default_pip_requirements())


def test_model_log_without_specified_conda_env_uses_default_env_with_expected_dependencies(
    reg_model,
):
    artifact_path = "model"
    with newron.start_run():
        newron.catboost.log_model(reg_model.model, artifact_path)
        model_uri = newron.get_artifact_uri(artifact_path)

    _assert_pip_requirements(model_uri, newron.catboost.get_default_pip_requirements())


def test_pyfunc_serve_and_score(reg_model):
    model, inference_dataframe = reg_model
    artifact_path = "model"
    with newron.start_run():
        newron.catboost.log_model(model, artifact_path)
        model_uri = newron.get_artifact_uri(artifact_path)

    resp = pyfunc_serve_and_score_model(
        model_uri,
        data=inference_dataframe,
        content_type=pyfunc_scoring_server.CONTENT_TYPE_JSON_SPLIT_ORIENTED,
        extra_args=EXTRA_PYFUNC_SERVING_TEST_ARGS,
    )
    scores = pd.read_json(resp.content.decode("utf-8"), orient="records").values.squeeze()
    np.testing.assert_array_almost_equal(scores, model.predict(inference_dataframe))


def test_pyfunc_serve_and_score_sklearn(reg_model):
    model, inference_dataframe = reg_model
    model = Pipeline([("model", reg_model.model)])

    with newron.start_run():
        newron.sklearn.log_model(model, artifact_path="model")
        model_uri = newron.get_artifact_uri("model")

    resp = pyfunc_serve_and_score_model(
        model_uri,
        inference_dataframe.head(3),
        pyfunc_scoring_server.CONTENT_TYPE_JSON_SPLIT_ORIENTED,
        extra_args=EXTRA_PYFUNC_SERVING_TEST_ARGS,
    )
    scores = pd.read_json(resp.content.decode("utf-8"), orient="records").values.squeeze()
    np.testing.assert_array_almost_equal(scores, model.predict(inference_dataframe.head(3)))


def test_log_model_with_code_paths(cb_model):
    artifact_path = "model"
    with newron.start_run(), mock.patch(
        "newron.catboost._add_code_from_conf_to_system_path"
    ) as add_mock:
        newron.catboost.log_model(cb_model.model, artifact_path, code_paths=[__file__])
        model_uri = newron.get_artifact_uri(artifact_path)
        _compare_logged_code_paths(__file__, model_uri, newron.catboost.FLAVOR_NAME)
        newron.catboost.load_model(model_uri=model_uri)
        add_mock.assert_called()


import os
from pathlib import Path

import pandas as pd

import mlflow
from mlflow import MlflowClient
from mlflow.utils.file_utils import read_yaml
from mlflow.pipelines.utils.execution import _MLFLOW_PIPELINES_EXECUTION_DIRECTORY_ENV_VAR
from mlflow.pipelines.utils import _PIPELINE_CONFIG_FILE_NAME
from mlflow.pipelines.steps.transform import TransformStep
from unittest import mock

# pylint: disable=unused-import
from tests.pipelines.helper_functions import tmp_pipeline_root_path

# pylint: enable=unused-import

# Sets up the transform step and returns the constructed TransformStep instance and step output dir
def set_up_transform_step(pipeline_root: Path, transform_user_module):
    split_step_output_dir = pipeline_root.joinpath("steps", "split", "outputs")
    split_step_output_dir.mkdir(parents=True)

    transform_step_output_dir = pipeline_root.joinpath("steps", "transform", "outputs")
    transform_step_output_dir.mkdir(parents=True)

    # use for train and validation, also for split
    dataset = pd.DataFrame(
        {
            "a": list(range(0, 5)),
            "b": list(range(5, 10)),
            "y": [float(i % 2) for i in range(5)],
        }
    )
    dataset.to_parquet(str(split_step_output_dir / "validation.parquet"))
    dataset.to_parquet(str(split_step_output_dir / "train.parquet"))

    pipeline_yaml = pipeline_root.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    experiment_name = "demo"
    MlflowClient().create_experiment(experiment_name)

    pipeline_yaml.write_text(
        """
        template: "regression/v1"
        target_col: "y"
        experiment:
          name: {experiment_name}
          tracking_uri: {tracking_uri}
        steps:
          transform:
            transformer_method: {transform_user_module}
        """.format(
            tracking_uri=mlflow.get_tracking_uri(),
            experiment_name=experiment_name,
            transform_user_module=transform_user_module,
        )
    )
    pipeline_config = read_yaml(pipeline_root, _PIPELINE_CONFIG_FILE_NAME)
    transform_step = TransformStep.from_pipeline_config(pipeline_config, str(pipeline_root))
    return transform_step, transform_step_output_dir, split_step_output_dir


def test_transform_step_writes_onehot_encoded_dataframe_and_transformer_pkl(tmp_pipeline_root_path):
    with mock.patch.dict(
        os.environ, {_MLFLOW_PIPELINES_EXECUTION_DIRECTORY_ENV_VAR: str(tmp_pipeline_root_path)}
    ):
        transform_step, transform_step_output_dir, _ = set_up_transform_step(
            tmp_pipeline_root_path, "sklearn.preprocessing.StandardScaler"
        )
        transform_step._run(str(transform_step_output_dir))

    assert os.path.exists(transform_step_output_dir / "transformed_training_data.parquet")
    transformed = pd.read_parquet(transform_step_output_dir / "transformed_training_data.parquet")
    assert len(transformed.columns) == 3
    assert os.path.exists(transform_step_output_dir / "transformer.pkl")


def test_transform_empty_step(tmp_pipeline_root_path):
    with mock.patch.dict(
        os.environ, {_MLFLOW_PIPELINES_EXECUTION_DIRECTORY_ENV_VAR: str(tmp_pipeline_root_path)}
    ), mock.patch("steps.transform.transformer_fn", return_value=None):
        transform_step, transform_step_output_dir, split_step_output_dir = set_up_transform_step(
            tmp_pipeline_root_path, "steps.transform.transformer_fn"
        )
        transform_step._run(str(transform_step_output_dir))

    assert os.path.exists(transform_step_output_dir / "transformed_training_data.parquet")
    train_transformed = pd.read_parquet(
        transform_step_output_dir / "transformed_training_data.parquet"
    )
    train_split = pd.read_parquet(split_step_output_dir / "train.parquet")

    assert train_transformed.equals(train_split) == True
    assert os.path.exists(transform_step_output_dir / "transformer.pkl")



import mlflow.spark
#### do we need classes to import e.g._PyFuncModelWrapper

FLAVOR_NAME = "spark"

# Default temporary directory on DFS. Used to write / read from Spark ML models.
DFS_TMP = "/tmp/newron"
_SPARK_MODEL_PATH_SUB = mlflow.spark._SPARK_MODEL_PATH_SUB
_MLFLOWDBFS_SCHEME = mlflow.spark._MLFLOWDBFS_SCHEME

get_default_pip_requirements = mlflow.spark.get_default_pip_requirements
get_default_conda_env = mlflow.spark.get_default_conda_env
log_model = mlflow.spark.log_model
_tmp_path = mlflow.spark._tmp_path
_mlflowdbfs_path = mlflow.spark._mlflowdbfs_path
_maybe_save_model = mlflow.spark._maybe_save_model
_should_use_mlflowdbfs = mlflow.spark._should_use_mlflowdbfs
_save_model_metadata = mlflow.spark._save_model_metadata
_validate_model = mlflow.spark._validate_model
save_model = mlflow.spark.save_model
_shutil_copytree_without_file_permissions = mlflow.spark._shutil_copytree_without_file_permissions
_load_model_databricks = mlflow.spark._load_model_databricks
_load_model = mlflow.spark._load_model
load_model = mlflow.spark.load_model
_load_pyfunc = mlflow.spark._load_pyfunc
_find_and_set_features_col_as_vector_if_needed = mlflow.spark._find_and_set_features_col_as_vector_if_needed
autolog = mlflow.spark.autolog


from unittest import mock
from pathlib import Path

import os
import pytest
import shutil
from sklearn.datasets import load_diabetes

import mlflow
from mlflow.utils.file_utils import read_yaml
from mlflow.pipelines.utils import _PIPELINE_CONFIG_FILE_NAME
from mlflow.pipelines.steps.split import _OUTPUT_TEST_FILE_NAME, _OUTPUT_VALIDATION_FILE_NAME
from mlflow.pipelines.steps.evaluate import EvaluateStep
from mlflow.exceptions import MlflowException

# pylint: disable=unused-import
from tests.pipelines.helper_functions import (
    clear_custom_metrics_module_cache,
    tmp_pipeline_exec_path,
    tmp_pipeline_root_path,
    train_and_log_model,
)  # pylint: enable=unused-import


@pytest.fixture(autouse=True)
def evaluation_inputs(tmp_pipeline_exec_path):
    split_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "split", "outputs")
    split_step_output_dir.mkdir(parents=True)
    X, y = load_diabetes(as_frame=True, return_X_y=True)
    validation_df = X.assign(y=y).sample(n=50, random_state=9)
    validation_df.to_parquet(split_step_output_dir.joinpath(_OUTPUT_VALIDATION_FILE_NAME))
    test_df = X.assign(y=y).sample(n=100, random_state=42)
    test_df.to_parquet(split_step_output_dir.joinpath(_OUTPUT_TEST_FILE_NAME))

    run_id, model = train_and_log_model()
    train_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "train", "outputs")
    train_step_output_dir.mkdir(parents=True)
    train_step_output_dir.joinpath("run_id").write_text(run_id)
    output_model_path = train_step_output_dir.joinpath("model")
    if os.path.exists(output_model_path) and os.path.isdir(output_model_path):
        shutil.rmtree(output_model_path)
    mlflow.sklearn.save_model(model, output_model_path)


@pytest.mark.usefixtures("clear_custom_metrics_module_cache")
@pytest.mark.parametrize("mae_threshold", [-1, 1_000_000])
def test_evaluate_step_run(
    tmp_pipeline_root_path: Path, tmp_pipeline_exec_path: Path, mae_threshold: int
):
    evaluate_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "evaluate", "outputs")
    evaluate_step_output_dir.mkdir(parents=True)

    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
    validation_criteria:
      - metric: root_mean_squared_error
        threshold: 1_000_000
      - metric: mean_absolute_error
        threshold: {mae_threshold}
      - metric: weighted_mean_squared_error
        threshold: 1_000_000
metrics:
  custom:
    - name: weighted_mean_squared_error
      function: weighted_mean_squared_error
      greater_is_better: False
""".format(
            tracking_uri=mlflow.get_tracking_uri(),
            mae_threshold=mae_threshold,
        )
    )
    pipeline_steps_dir = tmp_pipeline_root_path.joinpath("steps")
    pipeline_steps_dir.mkdir(parents=True)
    pipeline_steps_dir.joinpath("custom_metrics.py").write_text(
        """
def weighted_mean_squared_error(eval_df, builtin_metrics):
    from sklearn.metrics import mean_squared_error

    return {
        "weighted_mean_squared_error": mean_squared_error(
            eval_df["prediction"],
            eval_df["target"],
            sample_weight=1 / eval_df["prediction"].values,
        )
    }
"""
    )
    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    evaluate_step._run(str(evaluate_step_output_dir))

    logged_metrics = (
        mlflow.tracking.MlflowClient().get_run(mlflow.last_active_run().info.run_id).data.metrics
    )
    logged_metrics = {k.replace("_on_data_test", ""): v for k, v in logged_metrics.items()}
    assert "weighted_mean_squared_error" in logged_metrics
    model_validation_status_path = evaluate_step_output_dir.joinpath("model_validation_status")
    assert model_validation_status_path.exists()
    expected_status = "REJECTED" if mae_threshold < 0 else "VALIDATED"
    assert model_validation_status_path.read_text() == expected_status


@pytest.mark.usefixtures("clear_custom_metrics_module_cache")
def test_no_validation_criteria(tmp_pipeline_root_path: Path, tmp_pipeline_exec_path: Path):
    evaluate_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "evaluate", "outputs")
    evaluate_step_output_dir.mkdir(parents=True)

    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
""".format(
            tracking_uri=mlflow.get_tracking_uri()
        )
    )
    pipeline_steps_dir = tmp_pipeline_root_path.joinpath("steps")
    pipeline_steps_dir.mkdir(parents=True)
    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    evaluate_step._run(str(evaluate_step_output_dir))

    logged_metrics = (
        mlflow.tracking.MlflowClient().get_run(mlflow.last_active_run().info.run_id).data.metrics
    )
    logged_metrics = {k.replace("_on_data_test", ""): v for k, v in logged_metrics.items()}
    assert "mean_squared_error" in logged_metrics
    assert "root_mean_squared_error" in logged_metrics
    model_validation_status_path = evaluate_step_output_dir.joinpath("model_validation_status")
    assert model_validation_status_path.exists()
    assert model_validation_status_path.read_text() == "UNKNOWN"


@pytest.mark.usefixtures("clear_custom_metrics_module_cache", "tmp_pipeline_exec_path")
def test_validation_criteria_contain_undefined_metrics(tmp_pipeline_root_path: Path):
    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
    validation_criteria:
      - metric: root_mean_squared_error
        threshold: 100
      - metric: undefined_metric
        threshold: 100
""".format(
            tracking_uri=mlflow.get_tracking_uri()
        )
    )
    pipeline_steps_dir = tmp_pipeline_root_path.joinpath("steps")
    pipeline_steps_dir.mkdir(parents=True)

    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    with pytest.raises(
        MlflowException,
        match=r"Validation criteria contain undefined metrics: \['undefined_metric'\]",
    ):
        evaluate_step._validate_validation_criteria()


@pytest.mark.usefixtures("clear_custom_metrics_module_cache", "tmp_pipeline_exec_path")
def test_custom_metric_function_does_not_exist(
    tmp_pipeline_root_path: Path, tmp_pipeline_exec_path: Path
):
    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
    validation_criteria:
      - metric: weighted_mean_squared_error
        threshold: 100
metrics:
  custom:
    - name: weighted_mean_squared_error
      function: weighted_mean_squared_error
      greater_is_better: False
""".format(
            tracking_uri=mlflow.get_tracking_uri()
        )
    )
    pipeline_steps_dir = tmp_pipeline_root_path.joinpath("steps")
    pipeline_steps_dir.mkdir(parents=True)
    pipeline_steps_dir.joinpath("custom_metrics.py").write_text(
        """
def one(eval_df, builtin_metrics):
    return {"one": 1}
"""
    )
    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    evaluate_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "evaluate", "outputs")
    evaluate_step_output_dir.mkdir(parents=True)
    with pytest.raises(MlflowException, match="Failed to load custom metric functions") as exc:
        evaluate_step._run(str(evaluate_step_output_dir))
    assert isinstance(exc.value.__cause__, AttributeError)
    assert "weighted_mean_squared_error" in str(exc.value.__cause__)


@pytest.mark.usefixtures("clear_custom_metrics_module_cache", "tmp_pipeline_exec_path")
def test_custom_metrics_module_does_not_exist(
    tmp_pipeline_root_path: Path, tmp_pipeline_exec_path: Path
):
    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
    validation_criteria:
      - metric: weighted_mean_squared_error
        threshold: 100
metrics:
  custom:
    - name: weighted_mean_squared_error
      function: weighted_mean_squared_error
      greater_is_better: False
""".format(
            tracking_uri=mlflow.get_tracking_uri()
        )
    )
    pipeline_steps_dir = tmp_pipeline_root_path.joinpath("steps")
    pipeline_steps_dir.mkdir(parents=True)

    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    evaluate_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "evaluate", "outputs")
    evaluate_step_output_dir.mkdir(parents=True)
    with pytest.raises(MlflowException, match="Failed to load custom metric functions") as exc:
        evaluate_step._run(str(evaluate_step_output_dir))
    assert isinstance(exc.value.__cause__, ModuleNotFoundError)
    assert "No module named 'steps.custom_metrics'" in str(exc.value.__cause__)


@pytest.mark.usefixtures("clear_custom_metrics_module_cache", "tmp_pipeline_exec_path")
def test_custom_metrics_override_builtin_metrics(
    tmp_pipeline_root_path: Path, tmp_pipeline_exec_path: Path
):
    evaluate_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "evaluate", "outputs")
    evaluate_step_output_dir.mkdir(parents=True)

    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
    validation_criteria:
      - metric: root_mean_squared_error
        threshold: 10
      - metric: mean_absolute_error
        threshold: 10
metrics:
  custom:
    - name: mean_absolute_error
      function: mean_absolute_error
      greater_is_better: False
    - name: root_mean_squared_error
      function: root_mean_squared_error
      greater_is_better: False
""".format(
            tracking_uri=mlflow.get_tracking_uri()
        )
    )
    pipeline_steps_dir = tmp_pipeline_root_path.joinpath("steps")
    pipeline_steps_dir.mkdir(parents=True)
    pipeline_steps_dir.joinpath("custom_metrics.py").write_text(
        """
def mean_absolute_error(eval_df, builtin_metrics):
    return {"mean_absolute_error": 1}

def root_mean_squared_error(eval_df, builtin_metrics):
    return {"root_mean_squared_error": 1}
"""
    )
    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)

    with mock.patch("mlflow.pipelines.utils.metrics._logger.warning") as mock_warning:
        evaluate_step = EvaluateStep.from_pipeline_config(
            pipeline_config, str(tmp_pipeline_root_path)
        )
        evaluate_step._run(str(evaluate_step_output_dir))
        mock_warning.assert_called_once_with(
            "Custom metrics override the following built-in metrics: %s",
            ["mean_absolute_error", "root_mean_squared_error"],
        )
    logged_metrics = (
        mlflow.tracking.MlflowClient().get_run(mlflow.last_active_run().info.run_id).data.metrics
    )
    for dataset in ["validation", "test"]:
        assert f"root_mean_squared_error_on_data_{dataset}" in logged_metrics
        assert logged_metrics[f"root_mean_squared_error_on_data_{dataset}"] == 1
        assert f"mean_absolute_error_on_data_{dataset}" in logged_metrics
        assert logged_metrics[f"mean_absolute_error_on_data_{dataset}"] == 1
    model_validation_status_path = evaluate_step_output_dir.joinpath("model_validation_status")
    assert model_validation_status_path.exists()
    assert model_validation_status_path.read_text() == "VALIDATED"


def test_evaluate_step_writes_card_with_model_and_run_links_on_databricks(
    monkeypatch, tmp_pipeline_root_path: Path, tmp_pipeline_exec_path: Path
):
    workspace_host = "https://dev.databricks.com"
    workspace_id = 123456
    workspace_url = f"{workspace_host}?o={workspace_id}"

    monkeypatch.setenv("_DATABRICKS_WORKSPACE_HOST", workspace_host)
    monkeypatch.setenv("_DATABRICKS_WORKSPACE_ID", workspace_id)

    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
    validation_criteria:
      - metric: root_mean_squared_error
        threshold: 1_000_000
""".format(
            tracking_uri=mlflow.get_tracking_uri()
        )
    )

    evaluate_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "evaluate", "outputs")
    evaluate_step_output_dir.mkdir(parents=True)

    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    evaluate_step._run(str(evaluate_step_output_dir))

    train_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "train", "outputs")
    with open(train_step_output_dir / "run_id") as f:
        run_id = f.read()

    assert (evaluate_step_output_dir / "card.html").exists()
    with open(evaluate_step_output_dir / "card.html", "r") as f:
        step_card_content = f.read()

    assert f"<a href={workspace_url}#mlflow/experiments/1/runs/{run_id}>" in step_card_content
    assert (
        f"<a href={workspace_url}#mlflow/experiments/1/runs/{run_id}/artifactPath/train/model>"
        in step_card_content
    )



import pytest

import mlflow.pipelines.cli as pipelines_cli

from click.testing import CliRunner
from mlflow.exceptions import MlflowException
from mlflow.pipelines.utils import _PIPELINE_PROFILE_ENV_VAR

# pylint: disable=unused-import
from tests.pipelines.helper_functions import (
    enter_pipeline_example_directory,
)  # pylint: enable=unused-import

_STEP_NAMES = ["ingest", "split", "train", "transform", "evaluate", "register"]


@pytest.fixture
def clean_up_pipeline():
    try:
        yield
    finally:
        CliRunner().invoke(pipelines_cli.clean, env={_PIPELINE_PROFILE_ENV_VAR: "local"})


@pytest.mark.usefixtures("enter_pipeline_example_directory", "clean_up_pipeline")
@pytest.mark.parametrize("step", _STEP_NAMES)
def test_pipelines_cli_step_works(step):
    for command in [pipelines_cli.clean, pipelines_cli.inspect, pipelines_cli.run]:
        assert (
            CliRunner()
            .invoke(command, args=f"--step {step}", env={_PIPELINE_PROFILE_ENV_VAR: "local"})
            .exit_code
            == 0
        )
        assert CliRunner().invoke(command, args=f"--step {step} --profile=local").exit_code == 0


@pytest.mark.usefixtures("enter_pipeline_example_directory", "clean_up_pipeline")
def test_pipelines_cli_flow_completes_successfully():
    for command in [pipelines_cli.clean, pipelines_cli.inspect, pipelines_cli.run]:
        assert CliRunner().invoke(command, env={_PIPELINE_PROFILE_ENV_VAR: "local"}).exit_code == 0


@pytest.mark.usefixtures("enter_pipeline_example_directory", "clean_up_pipeline")
def test_pipelines_cli_fails_without_profile():
    for command in [pipelines_cli.clean, pipelines_cli.inspect, pipelines_cli.run]:
        assert CliRunner().invoke(command, env={_PIPELINE_PROFILE_ENV_VAR: ""}).exit_code != 0
        assert CliRunner().invoke(command).exit_code != 0


@pytest.mark.usefixtures("enter_pipeline_example_directory", "clean_up_pipeline")
def test_pipelines_cli_fails_with_illegal_profile():
    result = CliRunner().invoke(
        pipelines_cli.clean, env={_PIPELINE_PROFILE_ENV_VAR: "illegal_profile"}
    )
    assert result.exit_code != 0
    assert isinstance(result.exception, MlflowException)


@pytest.mark.usefixtures("enter_pipeline_example_directory", "clean_up_pipeline")
def test_pipelines_cli_works_with_non_default_profile():
    result = CliRunner().invoke(pipelines_cli.clean, env={_PIPELINE_PROFILE_ENV_VAR: "databricks"})
    assert "with profile: 'databricks'" in result.output
    assert result.exit_code == 0


@pytest.mark.usefixtures("enter_pipeline_example_directory", "clean_up_pipeline")
def test_pipelines_get_artifact_works():
    result = CliRunner().invoke(
        pipelines_cli.get_artifact,
        args="--artifact model",
        env={_PIPELINE_PROFILE_ENV_VAR: "local"},
    )
    assert result.exit_code == 0
    assert result.output != None


@pytest.mark.usefixtures("enter_pipeline_example_directory", "clean_up_pipeline")
def test_pipelines_get_artifact_with_bad_artifact_name_fails():
    result = CliRunner().invoke(
        pipelines_cli.get_artifact,
        args="--artifact foo",
        env={_PIPELINE_PROFILE_ENV_VAR: "local"},
    )
    assert result.exit_code != 0
    assert isinstance(result.exception, MlflowException)


@pytest.mark.usefixtures("enter_pipeline_example_directory", "clean_up_pipeline")
def test_pipelines_get_artifact_with_no_artifact_name_fails():
    result = CliRunner().invoke(
        pipelines_cli.get_artifact,
        env={_PIPELINE_PROFILE_ENV_VAR: "local"},
    )
    assert result.exit_code != 0
    assert isinstance(result.exception, SystemExit)



from mlflow.tracking.request_header.abstract_request_header_provider import RequestHeaderProvider


class PluginRequestHeaderProvider(RequestHeaderProvider):
    """RequestHeaderProvider provided through plugin system"""

    def in_context(self):
        return False

    def request_headers(self):
        return {"test": "header"}



import os
import pytest

import numpy as np
import pandas as pd

from mlflow.exceptions import MlflowException
from mlflow.pipelines.utils.execution import _MLFLOW_PIPELINES_EXECUTION_DIRECTORY_ENV_VAR
from mlflow.pipelines.steps.split import (
    _get_split_df,
    _hash_pandas_dataframe,
    _make_elem_hashable,
    SplitStep,
)
from unittest import mock


def test_split_step_run(tmp_path):
    ingest_output_dir = tmp_path / "steps" / "ingest" / "outputs"
    ingest_output_dir.mkdir(parents=True)
    split_output_dir = tmp_path / "steps" / "split" / "outputs"
    split_output_dir.mkdir(parents=True)

    num_rows = 1000
    num_good_rows = 900
    input_dataframe = pd.DataFrame(
        {
            "a": list(range(num_rows)),
            "b": [str(i) for i in range(num_rows)],
            "y": [float(i % 2) if i < num_good_rows else None for i in range(num_rows)],
        }
    )
    input_dataframe.to_parquet(str(ingest_output_dir / "dataset.parquet"))

    split_ratios = [0.6, 0.3, 0.1]

    with mock.patch.dict(
        os.environ, {_MLFLOW_PIPELINES_EXECUTION_DIRECTORY_ENV_VAR: str(tmp_path)}
    ), mock.patch("mlflow.pipelines.step.get_pipeline_name", return_value="fake_name"):
        split_step = SplitStep({"split_ratios": split_ratios, "target_col": "y"}, "fake_root")
        split_step._run(str(split_output_dir))

    (split_output_dir / "summary.html").exists()
    (split_output_dir / "card.html").exists()

    output_train_df = pd.read_parquet(str(split_output_dir / "train.parquet"))
    output_validation_df = pd.read_parquet(str(split_output_dir / "validation.parquet"))
    output_test_df = pd.read_parquet(str(split_output_dir / "test.parquet"))

    assert len(output_train_df) == 551
    assert len(output_validation_df) == 266
    assert len(output_test_df) == 83

    merged_output_df = pd.concat([output_train_df, output_validation_df, output_test_df])
    assert merged_output_df.columns.tolist() == ["a", "b", "y"]
    assert set(merged_output_df.a.tolist()) == set(range(num_good_rows))
    assert set(merged_output_df.b.tolist()) == set(str(i) for i in range(num_good_rows))
    assert set(merged_output_df.y.tolist()) == {0.0, 1.0}


def test_make_elem_hashable():
    assert _make_elem_hashable(3) == 3
    assert _make_elem_hashable("abc") == "abc"
    assert _make_elem_hashable((2, 3)) == (2, 3)
    assert _make_elem_hashable([2, 3]) == (2, 3)
    assert _make_elem_hashable([[2, 3], [4, 5]]) == ((2, 3), (4, 5))
    assert _make_elem_hashable({"a": 2, "b": 3}) == (("a", 2), ("b", 3))
    assert _make_elem_hashable({"a": [2, 3]}) == (("a", (2, 3)),)
    assert _make_elem_hashable(np.array([2, 3])) == ((2,), (2, 3))
    assert _make_elem_hashable(np.array([[2, 3, 4], [5, 6, 7]])) == ((2, 3), (2, 3, 4, 5, 6, 7))


def test_hash_pandas_dataframe_deterministic():
    pdf = pd.DataFrame(
        {
            "f1": [2, 3],
            "f2": [2.5, 3.5],
            "f3": [[6, 7], [8, 9]],
            "f4": [np.array([12, 13]), np.array([14, 15])],
            "f5": [np.array([12.5, 13.5]), np.array([14.5, 15.5])],
            "f6": [np.array([[22], [23]]), np.array([[24], [25]])],
            "f7": ["abc", "def"],
            "f8": [b"ghi", b"jkl"],
            "f9": [{"a1": 32, "b1": 33}, {"a2": 34, "b2": 35}],
            "f10": [{"a3": [42, 43]}, {"a4": [44, 45]}],
            "f11": pd.Series(["a5", "a6"], dtype="category"),
            "f13": [pd.Timestamp("2017-01-01T12"), pd.Timestamp("2017-02-01T11")],
            "f14": [True, False],
            "f15": [pd.Period("2000-01-02", freq="D"), pd.Period("2001-01-02", freq="M")],
            "f16": [pd.Interval(22.5, 23.5), pd.Interval(24.5, 25.5)],
        }
    )
    result = _hash_pandas_dataframe(pdf)
    assert result.tolist() == [
        2331111997514652279,
        12302536142759575339,
    ]


def test_get_split_df():
    with mock.patch("mlflow.pipelines.steps.split._SPLIT_HASH_BUCKET_NUM", 6):
        split_ratios = [3, 2, 1]
        hash_buckets = pd.Series([0.3, 0.9, 0.1, 0.7, 0.2, 0.6])
        dataset = pd.DataFrame({"v": [10, 20, 30, 40, 50, 60]})

        train_df, val_df, test_df = _get_split_df(dataset, hash_buckets, split_ratios)

        assert train_df.v.tolist() == [10, 30, 50]
        assert val_df.v.tolist() == [40, 60]
        assert test_df.v.tolist() == [20]


def test_from_pipeline_config_fails_without_target_col(tmp_path):
    with mock.patch.dict(
        os.environ, {_MLFLOW_PIPELINES_EXECUTION_DIRECTORY_ENV_VAR: str(tmp_path)}
    ), mock.patch(
        "mlflow.pipelines.step.get_pipeline_name", return_value="fake_name"
    ), pytest.raises(
        MlflowException, match="Missing target_col config"
    ):
        _ = SplitStep.from_pipeline_config({}, "fake_root")


def test_from_pipeline_config_works_with_target_col(tmp_path):
    with mock.patch.dict(
        os.environ, {_MLFLOW_PIPELINES_EXECUTION_DIRECTORY_ENV_VAR: str(tmp_path)}
    ), mock.patch("mlflow.pipelines.step.get_pipeline_name", return_value="fake_name"):
        assert SplitStep.from_pipeline_config({"target_col": "fake_col"}, "fake_root") is not None


def test_split_step_skips_profiling_when_specified(tmp_path):
    ingest_output_dir = tmp_path / "steps" / "ingest" / "outputs"
    ingest_output_dir.mkdir(parents=True)
    split_output_dir = tmp_path / "steps" / "split" / "outputs"
    split_output_dir.mkdir(parents=True)

    num_rows = 1000
    num_good_rows = 900
    input_dataframe = pd.DataFrame(
        {
            "a": list(range(num_rows)),
            "b": [str(i) for i in range(num_rows)],
            "y": [float(i % 2) if i < num_good_rows else None for i in range(num_rows)],
        }
    )
    input_dataframe.to_parquet(str(ingest_output_dir / "dataset.parquet"))

    with mock.patch.dict(
        os.environ, {_MLFLOW_PIPELINES_EXECUTION_DIRECTORY_ENV_VAR: str(tmp_path)}
    ), mock.patch(
        "mlflow.pipelines.utils.step.get_pandas_data_profile"
    ) as mock_profiling, mock.patch(
        "mlflow.pipelines.step.get_pipeline_name", return_value="fake_name"
    ):
        split_step = SplitStep({"target_col": "y", "skip_data_profiling": True}, "fake_root")
        split_step._run(str(split_output_dir))

    mock_profiling.assert_not_called()



import os
import subprocess
import requests
from concurrent.futures import ThreadPoolExecutor
import traceback


def get_distributions():
    res = subprocess.run(["pip", "list"], stdout=subprocess.PIPE, check=True)
    pip_list_stdout = res.stdout.decode("utf-8")
    # `pip_list_stdout` looks like this:
    # ``````````````````````````````````````````````````````````
    # Package     Version             Location
    # ----------- ------------------- --------------------------
    # mlflow      1.21.1.dev0         /home/user/path/to/mlflow
    # tensorflow  2.6.0
    # ...
    # ``````````````````````````````````````````````````````````
    lines = pip_list_stdout.splitlines()[2:]  # `[2:]` removes the header
    return [
        # Extract package and version
        line.split()[:2]
        for line in lines
    ]


def get_release_date(package, version):
    resp = requests.get(f"https://pypi.python.org/pypi/{package}/json", timeout=10)
    if not resp.ok:
        return ""

    matched = [dist_files for ver, dist_files in resp.json()["releases"].items() if ver == version]
    if (not matched) or (not matched[0]):
        return ""

    upload_time = matched[0][0]["upload_time"]
    return upload_time.split("T")[0]  # return year-month-day


def get_longest_string_length(array):
    return len(max(array, key=len))


def safe_result(future, if_error=""):
    try:
        return future.result()
    except Exception:
        traceback.print_exc()
        return if_error


def main():
    distributions = get_distributions()
    with ThreadPoolExecutor(max_workers=min(32, os.cpu_count() + 4)) as executor:
        futures = [executor.submit(get_release_date, pkg, ver) for pkg, ver in distributions]
        release_dates = [safe_result(f) for f in futures]

    packages, versions = list(zip(*distributions))
    package_legnth = get_longest_string_length(packages)
    version_length = get_longest_string_length(versions)
    release_date_length = len("Release Date")
    print("Package".ljust(package_legnth), "Version".ljust(version_length), "Release Date")
    print("-" * (package_legnth + version_length + release_date_length + 2))
    for package, version, release_date in sorted(
        zip(packages, versions, release_dates),
        # Sort by release date in descending order
        key=lambda x: x[2],
        reverse=True,
    ):
        print(
            package.ljust(package_legnth),
            version.ljust(version_length),
            release_date.ljust(release_date_length),
        )


if __name__ == "__main__":
    main()



import os
import pathlib
import shutil

import pytest

from mlflow.exceptions import MlflowException
from mlflow.pipelines.utils import (
    get_pipeline_root_path,
    get_pipeline_name,
    get_pipeline_config,
    get_default_profile,
)
from mlflow.utils.file_utils import write_yaml

# pylint: disable=unused-import
from tests.pipelines.helper_functions import (
    enter_pipeline_example_directory,
    enter_test_pipeline_directory,
)  # pylint: enable=unused-import
from tests.pipelines.helper_functions import chdir
from unittest import mock


def test_get_pipeline_root_path_returns_correctly_when_inside_pipeline_directory(
    enter_pipeline_example_directory,
):
    pipeline_root_path = enter_pipeline_example_directory
    assert get_pipeline_root_path() == str(pipeline_root_path)
    os.chdir(pathlib.Path.cwd() / "notebooks")
    assert get_pipeline_root_path() == str(enter_pipeline_example_directory)


def test_get_pipeline_root_path_throws_outside_pipeline_directory(tmp_path):
    with pytest.raises(MlflowException, match="Failed to find pipeline.yaml"), chdir(tmp_path):
        get_pipeline_root_path()


def test_get_pipeline_name_returns_correctly_for_valid_pipeline_directory(
    enter_pipeline_example_directory, tmp_path
):
    pipeline_root_path = enter_pipeline_example_directory
    assert pathlib.Path.cwd() == pipeline_root_path
    assert get_pipeline_name() == "sklearn_regression"

    with chdir(tmp_path):
        assert get_pipeline_name(pipeline_root_path=pipeline_root_path) == "sklearn_regression"


def test_get_pipeline_name_throws_for_invalid_pipeline_directory(tmp_path):
    with pytest.raises(MlflowException, match="Failed to find pipeline.yaml"), chdir(tmp_path):
        get_pipeline_name()

    with pytest.raises(MlflowException, match="Failed to find pipeline.yaml"):
        get_pipeline_name(pipeline_root_path=tmp_path)


def test_get_pipeline_config_returns_correctly_for_valid_pipeline_directory(
    enter_pipeline_example_directory, tmp_path
):
    pipeline_root_path = enter_pipeline_example_directory
    test_pipeline_root_path = tmp_path / "test_pipeline"
    shutil.copytree(pipeline_root_path, test_pipeline_root_path)

    test_pipeline_config = {
        "config1": 10,
        "config2": {
            "subconfig": ["A"],
        },
        "config3": "3",
    }
    write_yaml(test_pipeline_root_path, "pipeline.yaml", test_pipeline_config, overwrite=True)

    with chdir(test_pipeline_root_path):
        assert pathlib.Path.cwd() == test_pipeline_root_path
        assert get_pipeline_config() == test_pipeline_config

    with chdir(tmp_path):
        assert (
            get_pipeline_config(pipeline_root_path=test_pipeline_root_path) == test_pipeline_config
        )


def test_get_pipeline_config_throws_for_invalid_pipeline_directory(tmp_path):
    with pytest.raises(MlflowException, match="Failed to find pipeline.yaml"), chdir(tmp_path):
        get_pipeline_config()

    with pytest.raises(MlflowException, match="Failed to find pipeline.yaml"):
        get_pipeline_config(pipeline_root_path=tmp_path)


@pytest.mark.usefixtures("enter_pipeline_example_directory")
def test_get_pipeline_config_throws_for_nonexistent_profile():
    with pytest.raises(MlflowException, match="Did not find the YAML configuration.*badprofile"):
        get_pipeline_config(profile="badprofile")


def test_get_default_profile_works():
    assert get_default_profile() == "local"
    with mock.patch(
        "mlflow.pipelines.utils.is_in_databricks_runtime", return_value=True
    ) as patched_is_in_databricks_runtime:
        assert get_default_profile() == "databricks"
        patched_is_in_databricks_runtime.assert_called_once()


@pytest.mark.usefixtures("enter_test_pipeline_directory")
@pytest.mark.parametrize("get_config_from_subdirectory", [False, True])
def test_get_pipeline_config_succeeds_for_valid_profile(get_config_from_subdirectory):
    if get_config_from_subdirectory:
        with chdir("notebooks"):
            get_pipeline_config(profile="local")
    else:
        get_pipeline_config(profile="local")


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_get_pipeline_config_throws_clear_error_for_invalid_profile():
    with open("profiles/badcontent.yaml", "w") as f:
        f.write(r"\BAD")

    with pytest.raises(
        MlflowException,
        match="Failed to read pipeline configuration.*verify.*syntactically correct",
    ):
        get_pipeline_config(profile="badcontent")



# pylint: disable=arguments-differ
# pylint: disable=unused-argument
# pylint: disable=abstract-method

import pytorch_lightning as pl
import torch
from torch import nn
import torch.nn.functional as F

# NB: Older versions of PyTorch Lightning define native APIs for metric computation,
# (e.g., pytorch_lightning.metrics.Accuracy), while newer versions rely on the `torchmetrics`
# package (e.g. `torchmetrics.Accuracy)
try:
    from torchmetrics import Accuracy
except ImportError:
    from pytorch_lightning.metrics import Accuracy


class IrisClassificationBase(pl.LightningModule):
    def __init__(self, **kwargs):
        super().__init__()

        self.train_acc = Accuracy()
        self.val_acc = Accuracy()
        self.test_acc = Accuracy()
        self.args = kwargs

        self.fc1 = nn.Linear(4, 10)
        self.fc2 = nn.Linear(10, 10)
        self.fc3 = nn.Linear(10, 3)
        self.cross_entropy_loss = nn.CrossEntropyLoss()

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))
        return x

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), 0.01)


class IrisClassification(IrisClassificationBase):
    def training_step(self, batch, batch_idx):
        x, y = batch
        logits = self.forward(x)
        loss = self.cross_entropy_loss(logits, y)
        self.train_acc(torch.argmax(logits, dim=1), y)
        self.log("train_acc", self.train_acc.compute(), on_step=False, on_epoch=True)
        self.log("loss", loss)
        self.log("loss_forked", loss, on_epoch=True, on_step=True)
        return {"loss": loss}

    def validation_step(self, batch, batch_idx):
        x, y = batch
        logits = self.forward(x)
        loss = F.cross_entropy(logits, y)
        self.val_acc(torch.argmax(logits, dim=1), y)
        self.log("val_acc", self.val_acc.compute())
        self.log("val_loss", loss, sync_dist=True)

    def test_step(self, batch, batch_idx):
        x, y = batch
        logits = self.forward(x)
        loss = F.cross_entropy(logits, y)
        self.test_acc(torch.argmax(logits, dim=1), y)
        self.log("test_loss", loss)
        self.log("test_acc", self.test_acc.compute())


class IrisClassificationWithoutValidation(IrisClassificationBase):
    def training_step(self, batch, batch_idx):
        x, y = batch
        logits = self.forward(x)
        loss = self.cross_entropy_loss(logits, y)
        self.train_acc(torch.argmax(logits, dim=1), y)
        self.log("train_acc", self.train_acc.compute(), on_step=False, on_epoch=True)
        self.log("loss", loss)
        return {"loss": loss}

    def test_step(self, batch, batch_idx):
        x, y = batch
        logits = self.forward(x)
        loss = F.cross_entropy(logits, y)
        self.test_acc(torch.argmax(logits, dim=1), y)
        self.log("test_loss", loss)
        self.log("test_acc", self.test_acc.compute())


if __name__ == "__main__":
    pass


from setuptools import setup, find_packages


setup(
    name="mlflow-test-plugin",
    version="0.0.1",
    description="Test plugin for MLflow",
    packages=find_packages(),
    # Require MLflow as a dependency of the plugin, so that plugin users can simply install
    # the plugin & then immediately use it with MLflow
    install_requires=["mlflow"],
    entry_points={
        # Define a Tracking Store plugin for tracking URIs with scheme 'file-plugin'
        "mlflow.tracking_store": "file-plugin=mlflow_test_plugin.file_store:PluginFileStore",
        # Define a ArtifactRepository plugin for artifact URIs with scheme 'file-plugin'
        "mlflow.artifact_repository": "file-plugin=mlflow_test_plugin.local_artifact:PluginLocalArtifactRepository",  # pylint: disable=line-too-long
        # Define a RunContextProvider plugin. The entry point name for run context providers
        # is not used, and so is set to the string "unused" here
        "mlflow.run_context_provider": "unused=mlflow_test_plugin.run_context_provider:PluginRunContextProvider",  # pylint: disable=line-too-long
        # Define a DefaultExperimentProvider plugin. The entry point name for
        # default experiment providers is not used, and so is set to the string "unused" here
        "mlflow.default_experiment_provider": "unused=mlflow_test_plugin.default_experiment_provider:PluginDefaultExperimentProvider",  # pylint: disable=line-too-long
        # Define a RequestHeaderProvider plugin. The entry point name for request header providers
        # is not used, and so is set to the string "unused" here
        "mlflow.request_header_provider": "unused=mlflow_test_plugin.request_header_provider:PluginRequestHeaderProvider",  # pylint: disable=line-too-long
        # Define a Model Registry Store plugin for tracking URIs with scheme 'file-plugin'
        "mlflow.model_registry_store": "file-plugin=mlflow_test_plugin.sqlalchemy_store:PluginRegistrySqlAlchemyStore",  # pylint: disable=line-too-long
        # Define a MLflow Project Backend plugin called 'dummy-backend'
        "mlflow.project_backend": "dummy-backend=mlflow_test_plugin.dummy_backend:PluginDummyProjectBackend",  # pylint: disable=line-too-long
        # Define a MLflow model deployment plugin for target 'faketarget'
        "mlflow.deployments": "faketarget=mlflow_test_plugin.fake_deployment_plugin",
        # Define a Mlflow model evaluator with name "dummy_evaluator"
        "mlflow.model_evaluator": "dummy_evaluator=mlflow_test_plugin.dummy_evaluator:DummyEvaluator",  # pylint: disable=line-too-long
    },
)



"""
The ``newron.tensorflow`` module provides an API for logging and loading TensorFlow models.
"""

import mlflow.tensorflow
from packaging.version import Version
from newron.models import ModelSignature, ModelInputExample
from mlflow.utils.environment import (
    _mlflow_conda_env,
    _validate_env_arguments,
    _process_pip_requirements,
    _process_conda_env,
    _CONDA_ENV_FILE_NAME,
    _REQUIREMENTS_FILE_NAME,
    _CONSTRAINTS_FILE_NAME,
    _PYTHON_ENV_FILE_NAME,
    _PythonEnv,
)
from mlflow.utils.requirements_utils import _get_pinned_requirement
from collections import namedtuple
from threading import RLock
import inspect
FLAVOR_NAME = "tensorflow"
DEFAULT_AWAIT_MAX_SLEEP_SECONDS = 5 * 60
_TensorBoardLogDir = namedtuple("_TensorBoardLogDir", ["location", "is_temp"])

_load_pyfunc = mlflow.tensorflow._load_pyfunc
_validate_saved_model = mlflow.tensorflow._validate_saved_model
_load_tensorflow_saved_model = mlflow.tensorflow._load_tensorflow_saved_model
_parse_flavor_configuration = mlflow.tensorflow._parse_flavor_configuration
_TF2Wrapper = mlflow.tensorflow._TF2Wrapper
_assoc_list_to_map = mlflow.tensorflow._assoc_list_to_map
_flush_queue = mlflow.tensorflow._flush_queue
_add_to_queue = mlflow.tensorflow._add_to_queue
_add_to_queue = mlflow.tensorflow._add_to_queue
_get_tensorboard_callback = mlflow.tensorflow._get_tensorboard_callback
_setup_callbacks = mlflow.tensorflow._setup_callbacks
_metric_queue_lock = RLock()
_metric_queue = []


def get_default_pip_requirements():
    """
    Returns: 
            A list of default pip requirements for Models produced by Tensorflow.
             Calls to :func:`save_model()` and :func:`log_model()` produce a pip environment
             that, at minimum, contains these requirements.
    """
    import tensorflow as tf

    pip_deps = [_get_pinned_requirement("tensorflow")]

    # tensorflow >= 2.6.0 requires keras:
    # https://github.com/tensorflow/tensorflow/blob/v2.6.0/tensorflow/tools/pip_package/setup.py#L106
    # To prevent a different version of keras from being installed by tensorflow when creating
    # a serving environment, add a pinned requirement for keras
    if Version(tf.__version__) >= Version("2.6.0"):
        pip_deps.append(_get_pinned_requirement("keras"))

    return pip_deps

def get_default_conda_env():
    """
    Returns: 
            The default Conda environment for Models produced by calls to
             :func:`save_model()` and :func:`log_model()`.
    """
    return _mlflow_conda_env(additional_pip_deps=get_default_pip_requirements())

def log_model(
    tf_saved_model_dir,
    tf_meta_graph_tags,
    tf_signature_def_key,
    artifact_path,
    conda_env=None,
    code_paths=None,
    signature: ModelSignature = None,
    input_example: ModelInputExample = None,
    registered_model_name=None,
    await_registration_for=DEFAULT_AWAIT_MAX_SLEEP_SECONDS,
    pip_requirements=None,
    extra_pip_requirements=None,
):
    """
    Log a *serialized* collection of TensorFlow graphs and variables as an Newron model
    for the current run. This method operates on TensorFlow variables and graphs that have been
    serialized in TensorFlow's ``SavedModel`` format. For more information about ``SavedModel``
    format, see the TensorFlow documentation:
    [Tensorflow: Save and Restore Models](https://www.tensorflow.org/guide/saved_model#save_and_restore_models.)
    This method saves a model with both ``python_function`` and ``tensorflow`` flavors.
    If loaded back using the ``python_function`` flavor, the model can be used to predict on
    pandas DataFrames, producing a pandas DataFrame whose output columns correspond to the
    TensorFlow model's outputs. The python_function model will flatten outputs that are length-one,
    one-dimensional tensors of a single scalar value (e.g.
    ``{"predictions": [[1.0], [2.0], [3.0]]}``) into the scalar values (e.g.
    ``{"predictions": [1, 2, 3]}``), so that the resulting output column is a column of scalars
    rather than lists of length one. All other model output types are included as-is in the output
    DataFrame.
    Note that this method should not be used to log a ``tf.keras`` model. Use
    :py:func:`mlflow.keras.log_model` instead.

    Args:
        tf_saved_model_dir: Path to the directory containing serialized TensorFlow variables and
                               graphs in ``SavedModel`` format.
        tf_meta_graph_tags: A list of tags identifying the model's metagraph within the
                               serialized ``SavedModel`` object. For more information, see the
                               ``tags`` parameter of the
                               ``tf.saved_model.builder.SavedModelBuilder`` method.
        tf_signature_def_key: A string identifying the input/output signature associated with the
                                 model. This is a key within the serialized ``SavedModel`` signature
                                 definition mapping. For more information, see the
                                 ``signature_def_map`` parameter of the
                                 ``tf.saved_model.builder.SavedModelBuilder`` method.
        artifact_path: The run-relative path to which to log model artifacts.
        conda_env: conda_env
        code_paths: A list of local filesystem paths to Python file dependencies (or directories
                       containing file dependencies). These files are *prepended* to the system
                       path when the model is loaded.
        registered_model_name: If given, create a model version under
                                  ``registered_model_name``, also creating a registered model if one
                                  with the given name does not exist.
        signature: :py:class:`ModelSignature <mlflow.models.ModelSignature>`
                      describes model input and output :py:class:`Schema <mlflow.types.Schema>`.
                      The model signature can be :py:func:`inferred <mlflow.models.infer_signature>`
                      from datasets with valid model input (e.g. the training dataset with target
                      column omitted) and valid model output (e.g. model predictions generated on
                      the training dataset), for example:
                      .. code-block:: python
                        from mlflow.models.signature import infer_signature
                        train = df.drop_column("target_label")
                        predictions = ... # compute model predictions
                        signature = infer_signature(train, predictions)
        input_example: Input example provides one or several instances of valid
                          model input. The example can be used as a hint of what data to feed the
                          model. The given example can be a Pandas DataFrame where the given
                          example will be serialized to json using the Pandas split-oriented
                          format, or a numpy array where the example will be serialized to json
                          by converting it to a list. Bytes are base64-encoded.
        await_registration_for: Number of seconds to wait for the model version to finish
                            being created and is in ``READY`` status. By default, the function
                            waits for five minutes. Specify 0 or None to skip waiting.
        pip_requirements: pip_requirements
        extra_pip_requirements: extra_pip_requirements

    Returns: 
            A :py:class:`ModelInfo <mlflow.models.model.ModelInfo>` instance that contains the
             metadata of the logged model.
    """
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    del values["frame"]
    return mlflow.tensorflow.log_model(**values)
    
def save_model(
    tf_saved_model_dir,
    tf_meta_graph_tags,
    tf_signature_def_key,
    path,
    mlflow_model=None,
    conda_env=None,
    code_paths=None,
    signature: ModelSignature = None,
    input_example: ModelInputExample = None,
    pip_requirements=None,
    extra_pip_requirements=None,
):
    """
    Save a *serialized* collection of TensorFlow graphs and variables as an MLflow model
    to a local path. This method operates on TensorFlow variables and graphs that have been
    serialized in TensorFlow's ``SavedModel`` format. For more information about ``SavedModel``
    format, see the TensorFlow documentation:
    [Tensorflow: Save and Restore Models](https://www.tensorflow.org/guide/saved_model#save_and_restore_models.)

    Args:[source][source][source]
        tf_saved_model_dir: Path to the directory containing serialized TensorFlow variables and
                               graphs in ``SavedModel`` format.
        tf_meta_graph_tags: A list of tags identifying the model's metagraph within the
                               serialized ``SavedModel`` object. For more information, see the
                               ``tags`` parameter of the
                               ``tf.saved_model.builder.savedmodelbuilder`` method.
        tf_signature_def_key: A string identifying the input/output signature associated with the
                                 model. This is a key within the serialized ``savedmodel``
                                 signature definition mapping. For more information, see the
                                 ``signature_def_map`` parameter of the
                                 ``tf.saved_model.builder.savedmodelbuilder`` method.
        path: Local path where the MLflow model is to be saved.
        mlflow_model: MLflow model configuration to which to add the ``tensorflow`` flavor.
        conda_env: conda_env
        code_paths: A list of local filesystem paths to Python file dependencies (or directories
                       containing file dependencies). These files are *prepended* to the system
                       path when the model is loaded.
        signature: :py:class:`ModelSignature <mlflow.models.ModelSignature>`
                      describes model input and output :py:class:`Schema <mlflow.types.Schema>`.
                      The model signature can be :py:func:`inferred <mlflow.models.infer_signature>`
                      from datasets with valid model input (e.g. the training dataset with target
                      column omitted) and valid model output (e.g. model predictions generated on
                      the training dataset), for example:
                      .. code-block:: python
                        from newron.models.signature import infer_signature
                        train = df.drop_column("target_label")
                        predictions = ... # compute model predictions
                        signature = infer_signature(train, predictions)
        input_example: Input example provides one or several instances of valid
                          model input. The example can be used as a hint of what data to feed the
                          model. The given example can be a Pandas DataFrame where the given
                          example will be serialized to json using the Pandas split-oriented
                          format, or a numpy array where the example will be serialized to json
                          by converting it to a list. Bytes are base64-encoded.
        pip_requirements: pip_requirements
        extra_pip_requirements: extra_pip_requirements
    """
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    del values["frame"]
    mlflow.tensorflow.save_model(**values)

def load_model(model_uri, dst_path=None):
    """
    Load an Newron model that contains the TensorFlow flavor from the specified path.
    ```python
        import newron.tensorflow
        import tensorflow as tf
        tf_graph = tf.Graph()
        tf_sess = tf.Session(graph=tf_graph)
        with tf_graph.as_default():
            signature_definition = newron.tensorflow.load_model(model_uri="model_uri",
                                    tf_sess=tf_sess)
            input_tensors = [tf_graph.get_tensor_by_name(input_signature.name)
                                for _, input_signature in signature_definition.inputs.items()]
            output_tensors = [tf_graph.get_tensor_by_name(output_signature.name)
                                for _, output_signature in signature_definition.outputs.items()]
    ```
    Args:
        model_uri: The location, in URI format, of the Newron model. For example:
            * ``Users/me/path/to/local/model``
            * ``relative/path/to/local/model``
            * ``s3://my_bucket/path/to/model``
            * ``models:/<model_name>/<model_version>``
            * ``models:/<model_name>/<stage>``
                    For more information about supported URI schemes, see
                    `Referencing Artifacts <https://www.mlflow.org/docs/latest/concepts.html#
                    artifact-locations>`.
        dst_path: The local filesystem path to which to download the model artifact.
                     This directory must already exist. If unspecified, a local output
                     path will be created.
    Returns: 
        A callable graph (tf.function) that takes inputs and returns inferences.   
    """
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    del values["frame"]
    return mlflow.tensorflow.load_model(**values)

def autolog(
    every_n_iter=1,
    log_models=True,
    disable=False,
    exclusive=False,
    disable_for_unsupported_versions=False,
    silent=False,
    registered_model_name=None,
    log_input_examples=False,
    log_model_signatures=False,
):  # pylint: disable=unused-argument
    # pylint: disable=E0611
    """
    Enables automatic logging from TensorFlow to Newron.
    Note that autologging for ``tf.keras`` is handled by :py:func:`mlflow.tensorflow.autolog`,
    not :py:func:`mlflow.keras.autolog`.
    As an example, try running the
    `TensorFlow examples <https://github.com/mlflow/mlflow/tree/master/examples/tensorflow>`.
    For each TensorFlow module, autologging captures the following information:
    **tf.keras**
     - **Metrics** and **Parameters**
      - Training loss; validation loss; user-specified metrics
      - ``fit()`` or ``fit_generator()`` parameters; optimizer name; learning rate; epsilon
     - **Artifacts**
      - Model summary on training start
      - `MLflow Model <https://mlflow.org/docs/latest/models.html>` (Keras model)
      - TensorBoard logs on training end
    **tf.keras.callbacks.EarlyStopping**
     - **Metrics** and **Parameters**
      - Metrics from the ``EarlyStopping`` callbacks: ``stopped_epoch``, ``restored_epoch``,``restore_best_weight``, etc
      - ``fit()`` or ``fit_generator()`` parameters associated with ``EarlyStopping``:
        ``min_delta``, ``patience``, ``baseline``, ``restore_best_weights``, etc
    **tf.estimator**
     - **Metrics** and **Parameters**
      - TensorBoard metrics: ``average_loss``, ``loss``, etc
      - Parameters ``steps`` and ``max_steps``
     - **Artifacts**
      - `Newron Model <https://newron.org/docs/latest/models.html>` (TF saved model) on call
        to ``tf.estimator.export_saved_model``
    **TensorFlow Core**
     - **Metrics**
      - All ``tf.summary.scalar`` calls
    Refer to the autologging tracking documentation for more
    information on `TensorFlow workflows <https://www.mlflow.org/docs/latest/tracking.html#tensorflow-and-keras-experimental>`.
    
    Args:
        every_n_iter: The frequency with which metrics should be logged. For example, a value of
                         100 will log metrics at step 0, 100, 200, etc.
        log_models: If ``True``, trained models are logged as Newron model artifacts.
                       If ``False``, trained models are not logged.
        disable: If ``True``, disables the TensorFlow autologging integration. If ``False``,
                    enables the TensorFlow integration autologging integration.
        exclusive: If ``True``, autologged content is not logged to user-created fluent runs.
                      If ``False``, autologged content is logged to the active fluent run,
                      which may be user-created.
        disable_for_unsupported_versions: If ``True``, disable autologging for versions of
                      tensorflow that have not been tested against this version of the Newron
                      client or are incompatible.
        silent: If ``True``, suppress all event logs and warnings from Newron during TensorFlow
                   autologging. If ``False``, show all events and warnings during TensorFlow
                   autologging.
        registered_model_name: If given, each time a model is trained, it is registered as a
                                  new model version of the registered model with this name.
                                  The registered model is created if it does not already exist.
        log_input_examples: If ``True``, input examples from training datasets are collected and
                               logged along with tf/keras model artifacts during training. If
                               ``False``, input examples are not logged.
        log_model_signatures: If ``True``,
                                 :py:class:`ModelSignatures <mlflow.models.ModelSignature>`
                                 describing model inputs and outputs are collected and logged along
                                 with tf/keras model artifacts during training. If ``False``,
                                 signatures are not logged. ``False`` by default because
                                 logging TensorFlow models with signatures changes their pyfunc
                                 inference behavior when Pandas DataFrames are passed to
                                 ``predict()``: when a signature is present, an ``np.ndarray``
                                 (for single-output models) or a mapping from
                                 ``str`` -> ``np.ndarray`` (for multi-output models) is returned;
                                 when a signature is not present, a Pandas DataFrame is returned.
    """

    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    del values["frame"]
    mlflow.tensorflow.autolog(**values)